from .average_recommender import AverageRecommender
from .random_recommender import RandomRecommender
from .knn_meta_recommender import KNNMetaRecommender
from .surprise_recommenders import (
    CoClusteringRecommender,
    KNNWithMeansRecommender,
    KNNDatasetRecommender,
    KNNMLRecommender,
    SlopeOneRecommender,
    SVDRecommender)



import urllib.request, urllib.parse
import json
import os
import datetime 
unixtime = int(datetime.datetime.now().strftime("%s")) * 1000
#last 8 hours
lastupdate = unixtime - (8 * 60 * 60 * 1000 )
#baseURL = os.environ['FGLAB_URL']
baseURL = 'http://hoth.pmacs.upenn.edu:5080'
experimentsURL=baseURL+'/api/experiments'
#postvars = {'apikey':os.environ['APIKEY'],'date_start':lastupdate} 
postvars = {'apikey':'Oed+kIyprDrUq/3oWU5Jpyd22PqhG/CsUvI8oc9l39E=','date_start':lastupdate}
params = json.dumps(postvars).encode('utf8')
req = urllib.request.Request(experimentsURL, data=params,
  headers={'content-type': 'application/json'})
r = urllib.request.urlopen(req)
data = json.loads(r.read().decode(r.info().get_param('charset') or 'utf-8'))
print(len(data))



from sklearn.tree import DecisionTreeClassifier
import numpy as np
import pandas as pd
from time import sleep
import time
import datetime
import pickle
import pdb
import ai.api_utils as api_utils
from ai.api_utils import LabApi, AI_STATUS, RECOMMENDER_STATUS
import ai.q_utils as q_utils
import os
import ai.knowledgebase_utils as knowledgebase_utils
import logging
import sys
from ai.recommender.average_recommender import AverageRecommender
from ai.recommender.random_recommender import RandomRecommender
from ai.recommender.knn_meta_recommender import KNNMetaRecommender
# from ai.recommender.svd_recommender import SVDRecommender
from ai.recommender.surprise_recommenders import (CoClusteringRecommender,
        KNNWithMeansRecommender, KNNDatasetRecommender, KNNMLRecommender,
        SlopeOneRecommender, SVDRecommender)

from ai.request_manager import RequestManager

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
formatter = logging.Formatter('%(module)s: %(levelname)s: %(message)s')
ch.setFormatter(formatter)
logger.addHandler(ch)




class AI():
    """AI managing agent for Penn AI.

    Responsible for:
    - checking for user requests for recommendations,
    - checking for new results from experiments,
    - calling the recommender system to generate experiment recommendations,
    - posting the recommendations to the API.
    - handling communication with the API.

    :param rec_class: ai.BaseRecommender - recommender to use
    :param api_path: string - path to the lab api server
    :param extra_payload: dict - any additional payload that needs to be 
    specified
    :param user: string - test user
    :param rec_score_file: file - pickled score file to keep persistent scores
    between sessions
    :param verbose: Boolean
    :param warm_start: Boolean - if true, attempt to load the ai state from the 
    file provided by rec_score_file
    :param n_recs: int - number of recommendations to make for each request
    :param datasets: str or False - if not false, a comma seperated list of 
    datasets to turn the ai on for at startup
    :param use_pmlb_knowledgebase: Boolean

    """

    def __init__(self,
                rec_class=None,
                api_path=None,
                extra_payload=dict(),
                user='testuser',
                rec_score_file='rec_state.obj',
                verbose=True,
                warm_start=False,
                n_recs=1,
                datasets=False,
                use_knowledgebase=False,
                term_condition='n_recs',
                max_time=5):
        """Initializes AI managing agent."""
        if 'RANDOM_SEED' in os.environ:
            self.random_state = int(os.environ['RANDOM_SEED'])
        else:
            self.random_state = 0
        # default supervised learning recommender settings
        self.DEFAULT_REC_CLASS = {'classification':RandomRecommender,
                                  'regression':RandomRecommender
                                  }
        self.DEFAULT_REC_ARGS = {
                'classification': {
                    'ml_type':'classifier',
                    'metric':'accuracy',
                    'random_state':self.random_state
                    },
                'regression': {
                    'ml_type':'regressor',
                    'metric':'r2_cv_mean',
                    'random_state':self.random_state
                    }
                }

        # recommendation engines for different problem types
        # will be expanded as more types of probles are supported
        # (classification, regression, unsupervised, etc.)
        self.rec_engines = {
            "classification":None,
            "regression":None
        }

        # Request manager settings
        self.n_recs=n_recs if n_recs>0 else 1
        self.continuous= n_recs<1

        # api parameters, will be removed from self once the recommenders no 
        # longer call the api directly.
        # See #98 <https://github.com/EpistasisLab/pennai/issues/98>
        if api_path == None:
            api_path = ('http://' + os.environ['LAB_HOST'] + ':' +
                        os.environ['LAB_PORT'])
        self.user=user
        self.api_path=api_path
        self.api_key=os.environ['APIKEY']

        self.verbose = verbose #False: no printouts, True: printouts on updates

        # file name of stored scores for the recommender
        self.rec_score_file = rec_score_file

        # timestamp of the last time new experiments were processed
        self.last_update = 0

        # api
        self.labApi = api_utils.LabApi(
            api_path=self.api_path,
            user=self.user,
            api_key=self.api_key,
            extra_payload=extra_payload,
            verbose=self.verbose)

        
        # build dictionary of ml ids to names conversion
        self.ml_id_to_name = self.labApi.get_ml_id_dict()
        self.ml_name_to_id = {v:k for k,v in self.ml_id_to_name.items()}
        # print('ml_id_to_name:',self.ml_id_to_name)

        # dictionary of dataset threads, initialized and used by q_utils.
        # Keys are datasetIds, values are q_utils.DatasetThread instances.
        self.dataset_threads = {}

        # local dataframe of datasets and their metafeatures
        self.dataset_mf_cache = pd.DataFrame()

        # store dataset_id to hash dictionary
        self.dataset_mf_cache_id_hash_lookup = {}

        # set recommender status
        self.labApi.set_recommender_status(
                RECOMMENDER_STATUS.INITIALIZING.value)

        #initialize recommenders
        self.use_knowledgebase = use_knowledgebase
        self.initialize_recommenders(rec_class) # set self.rec_engines

        # set termination condition
        self.term_condition = term_condition
        if self.term_condition == 'n_recs':
            self.term_value = self.n_recs
        elif self.term_condition == 'time':
            self.term_value = max_time
        else:
            self.term_value = None

        # start the request manager
        self.requestManager = RequestManager(
            ai=self,
            defaultTermConditionStr=self.term_condition,
            defaultTermParam=self.term_value)

        # if there is a pickle file, load it as the recommender scores
        assert not (warm_start), "The `warm_start` option is not yet supported"

        # for comma-separated list of datasets in datasets, turn AI request on
        assert not datasets, \
                "The `datasets` option is not yet supported: " + str(datasets)

        # set recommender status
        self.labApi.set_recommender_status(RECOMMENDER_STATUS.RUNNING.value)

    ##-----------------
    ## Init methods
    ##-----------------
    def initialize_recommenders(self, rec_class):
        """
        Initilize classification and regression recommenders
        """

        kb = None
        if self.use_knowledgebase:
            kb = self.load_knowledgebase()

        for pred_type in self.rec_engines.keys():
            logger.info('initialiazing rec engine for problem type "'
                    +pred_type+'"')

            # get the ml parameters for the given recommender type
            logger.debug("getting ml_p")
            ml_p = self.labApi.get_all_ml_p(pred_type)
            assert ml_p is not None
            assert len(ml_p) > 0

            # Create supervised learning recommenders
            logger.debug("initializing engine")
            recArgs = self.DEFAULT_REC_ARGS[pred_type]
            recArgs['ml_p'] = ml_p

            recArgs['serialized_rec_directory'] = 'data/recommenders/pennaiweb'
            recArgs['load_serialized_rec'] = "if_exists" 

            if kb is not None:
                recArgs['knowledgebase_results'] = kb['resultsData'][pred_type]

            if (rec_class):
                self.rec_engines[pred_type] = rec_class(**recArgs)
            else:
                self.rec_engines[pred_type]  = \
                        self.DEFAULT_REC_CLASS[pred_type](**recArgs)


            # self.rec_engines[pred_type].update(kb['resultsData'][pred_type], 
            #         self.dataset_mf_cache, source='knowledgebase')
            ##########################################################
            # this section is used to save trained recommenders
            # on the PMLB knowledgebases.            
            # For normal operation, they can be skipped.
            # logger.info('saving recommender')
            # self.rec_engines[pred_type].save()
            ##########################################################

        logger.debug("recomendation engines initialized: ")
        for prob_type, rec in self.rec_engines.items():
            logger.debug(f'\tproblemType: {prob_type} - {rec}')
            #logger.debug('\trec.ml_p:\n'+str(rec.ml_p.head()))


    def load_knowledgebase(self):
        """Bootstrap the recommenders with the knowledgebase."""
        logger.info('loading pmlb knowledgebase')

        kb = knowledgebase_utils.load_default_knowledgebases(dedupe=False)
        
        all_df_mf = kb['metafeaturesData'].set_index('_id')

        # replace algorithm names with their ids
        
        for pred_type in ['classification','regression']:
            kb['resultsData'][pred_type]['algorithm'] = \
                    kb['resultsData'][pred_type]['algorithm'].apply(
                      lambda x: x
                      if x in self.ml_name_to_id.keys() 
                      else 'REMOVE! ' + x)
            # filter any kb results that we don't have an algorithm for
            for algs in kb['resultsData'][pred_type]['algorithm'].unique():
                if 'REMOVE!' in algs:
                    logger.warn('Removing knowledgebase results for algorithm '
                            +algs[7:]
                            +' because that algorithm is not available')
            kb['resultsData'][pred_type] = kb['resultsData'][pred_type].loc[
                    ~(kb['resultsData'][pred_type]['algorithm'].str.contains(
                        'REMOVE!'))
                    ]

            # use _id to index the metafeatures, and
            # keep only metafeatures with results
            self.dataset_mf_cache.append(
                all_df_mf.loc[kb['resultsData'][pred_type]['_id'].unique()]
                    )

            logger.info(f"updating AI with {pred_type} knowledgebase ("
                f"{len(kb['resultsData'][pred_type])} results)")

            logger.info('pmlb '+pred_type+' knowledgebase loaded')

        return kb


    ##-----------------
    ## Utility methods
    ##-----------------
    def get_results_metafeatures(self, results_data):
        """
        Return a pandas dataframe of metafeatures associated with the datasets
        in results_data.

        Retireves metafeatures from self.dataset_mf_cache if they exist,
        otherwise queries the api and updates the cache.

        :param results_data: experiment results with associated datasets

        """
        #logger.debug('results_data:'+str(results_data.columns))
        #logger.debug('results_data:'+str(results_data.head()))

        dataset_metafeatures = []
        
        dataset_indices = results_data['dataset_id'].unique()

        # add dataset metafeatures to the cache
        for d in dataset_indices:
            if (len(self.dataset_mf_cache)==0 
                or d not in self.dataset_mf_cache_id_hash_lookup.keys()):
                df = self.labApi.get_metafeatures(d)        
                df['dataset'] = d
                dataset_metafeatures.append(df)
                self.dataset_mf_cache_id_hash_lookup.update({d:df['_id']})
        if dataset_metafeatures:
            df_mf = pd.concat(dataset_metafeatures).set_index('dataset')
            self.dataset_mf_cache = self.dataset_mf_cache.append(df_mf)


        logger.debug(f'mf count:\n {len(self.dataset_mf_cache.index.values)}')
        #logger.info(f'mf:\n {list(self.dataset_mf_cache.index.values)}')
        logger.info(f'indices: \n\n {dataset_indices}')

        new_mf = self.dataset_mf_cache.loc[dataset_indices, :]
        assert len(new_mf) == len(dataset_indices)
        #logger.debug(f"new_mf: {new_mf}")

        return new_mf

    ##-----------------
    ## Loop methods
    ##-----------------
    def check_results(self):
        """Checks to see if new experiment results have been posted since the
        previous time step. If so, set them to self.new_data and return True.

        :returns: Boolean - True if new results were found
        """
        logger.info(time.strftime("%Y %I:%M:%S %p %Z",time.localtime())+
                  ': checking results...')

        newResults = self.labApi.get_new_experiments_as_dataframe(
                                        last_update=self.last_update)

        if len(newResults) > 0:
            newResults['algorithm'] = newResults['algorithm_id'].apply(
                lambda x: self.ml_id_to_name[x])
            logger.info(time.strftime("%Y %I:%M:%S %p %Z",time.localtime())+
                  ': ' + str(len(newResults)) + ' new results!')
            self.last_update = int(time.time())*1000 # update timestamp
            self.new_data = newResults
            return True

        return False

    def update_recommender(self):
        """Update recommender models based on new experiment results in
        self.new_data, and then clear self.new_data.
        """
        if(hasattr(self,'new_data') and len(self.new_data) >= 1):
            for predictionType in self.new_data.prediction_type.unique():
                filterRes = self.new_data[
                        self.new_data['prediction_type']==predictionType]
                filterMf = self.get_results_metafeatures(filterRes)

                filterRes['_id'] = filterRes['dataset_id'].apply(
                    lambda x: self.dataset_mf_cache_id_hash_lookup[x])


                self.rec_engines[predictionType].update(filterRes, filterMf)
                logger.info(
                        time.strftime("%Y %I:%M:%S %p %Z",time.localtime())+
                    ': {} recommender updated with '
                    '{} result(s)'.format(predictionType, 
                        len(filterRes.index)))

            # reset new data
            self.new_data = pd.DataFrame()

    def check_requests(self):
        """Check to see if any new AI requests have been submitted.
        If so, add them to self.request_queue.

        :returns: Boolean - True if new AI requests have been submitted
        """

        logger.info(time.strftime("%Y %I:%M:%S %p %Z",time.localtime())+
              ': checking requests...')


        # get all dtasets that have an ai 'requested' status
        # and initialize a new request
        dsFilter = {'ai':[AI_STATUS.REQUESTED.value, 'dummy']}
        aiOnRequests = self.labApi.get_filtered_datasets(dsFilter)

        if len(aiOnRequests) > 0:
            logger.info(time.strftime("%Y %I:%M:%S %p %Z",time.localtime())+
                      ': new ai request for:'+
                      ';'.join([r['name'] for r in aiOnRequests]))

            # set AI flag to 'on' to acknowledge requests received
            for r in aiOnRequests:
                self.labApi.set_ai_status(datasetId = r['_id'],
                                            aiStatus = 'on')

                self.requestManager.add_request(
                        datasetId=r['_id'], 
                        datasetName=r['name']) 
        time.sleep(.1)
        # get all datasets that have a manual 'off' status
        # and terminate their ai requests
        dsFilter = {'ai':[AI_STATUS.OFF.value, 'dummy']}
        aiOffRequests = self.labApi.get_filtered_datasets(dsFilter)

        if len(aiOffRequests) > 0:
            logger.info(time.strftime("%Y %I:%M:%S %p %Z",time.localtime())+
                      ': ai termination request for:'+
                      ';'.join([r['name'] for r in aiOffRequests]))

            for r in aiOffRequests:
                self.requestManager.terminate_request(datasetId=r['_id'])

        return True

    def process_rec(self):
        self.requestManager.process_requests()

    ##-----------------
    ## Syncronous actions an AI request can take
    ##-----------------
    def generate_recommendations(self, datasetId, numOfRecs):
        """Generate ml recommendation payloads for the given dataset.

        :param datasetId
        :param numOfRecs

        :returns list of maps that represent request payload objects
        """
        logger.info(f'generate_recommendations({datasetId},{numOfRecs}')

        recommendations = []

        metafeatures = self.labApi.get_metafeatures(datasetId)
        assert '_prediction_type' in metafeatures.columns
        predictionType = metafeatures['_prediction_type'].values[0]

        logger.info("prediction_type: " + predictionType)


        ml, p, ai_scores = self.rec_engines[predictionType].recommend(
            dataset_id=metafeatures['_id'].values[0],
            n_recs=numOfRecs,
            dataset_mf=metafeatures)

        for alg,params,score in zip(ml,p,ai_scores):
            # TODO: just return dictionaries of parameters from rec
            # modified_params = eval(params) # turn params into a dictionary

            recommendations.append({'dataset_id':datasetId,
                    'algorithm_id':self.ml_name_to_id[alg],
                    'username':self.user,
                    'parameters':params,
                    'ai_score':score,
                    })

        return recommendations


    def transfer_rec(self, rec_payload):
        """Attempt to send a recommendation to the lab server.
        If any error other then a no capacity error occurs, throw an exception.

        :param rec_payload: dictionary - the payload describing the experiment

        :return bool - true if successfully sent, false if no machine capacity 
        available
        """
        logger.info(f"transfer_rec({rec_payload})")

        aiStatus = self.labApi.get_dataset_ai_status(rec_payload['dataset_id'])
        if not(aiStatus == AI_STATUS.ON.value):
            logger.debug("AI status is not on; not submitting experiment")
            return False

        submitstatus = self.labApi.launch_experiment(
                            algorithmId=rec_payload['algorithm_id'],
                            payload=rec_payload)

        logger.debug(f"transfer_rec() submitstatus: {submitstatus}")

        if 'error' in submitstatus:
            if ('No machine capacity available' in submitstatus['error']):
                logger.debug(f"Waiting for capacity: {submitstatus['error']}")
                return False
            else:
                msg = ('Unrecoverable error during transfer_rec : ' 
                        + str(submitstatus))
                logger.error(msg)
                raise RuntimeError(msg)

        return True


    ##-----------------
    ## Save/load ai state
    ##-----------------
    def save_state(self):
        """Save ML+P scores in pickle or to DB

        TODO: test that this still works
        """
        raise RuntimeError("save_state is not currently supported")
        out = open(self.rec_score_file,'wb')
        state={}
        if(hasattr(self.rec_engines["classification"], 'scores')):
            #TODO: make this a more generic. Maybe just save the
            # AI or rec object itself.
            state['scores'] = self.rec_engines["classification"].scores
            state['last_update'] = self.last_update
        pickle.dump(state, out)

    def load_state(self):
        """Loads pickled score file and recommender model.

        TODO: test that this still works
        """
        raise RuntimeError("load_state is not currently supported")
        if os.stat(self.rec_score_file).st_size != 0:
            filehandler = open(self.rec_score_file,'rb')
            state = pickle.load(filehandler)
            if(hasattr(self.rec_engines["classification"], 'scores')):
              self.rec_engines["classification"].scores = state['scores']
              self.last_update = state['last_update']
              logger.info('loaded previous state from '+self.last_update)



####################################################################### Manager
import argparse

def main():
    """Handles command line arguments and runs Penn-AI."""
    parser = argparse.ArgumentParser(
            description='PennAI is a recommendation system for data science. ',
            add_help=False)
    parser.add_argument('-h','--help',action='help',
                        help="Show this help message and exit.")
    parser.add_argument('-rec',action='store',dest='REC',default='random',
            choices = ['random','average','knnmeta','svd','cocluster',
                'knnmeans', 'knnml','knndata','slopeone'],
            help='Recommender algorithm options.')
    parser.add_argument('-api_path',action='store',dest='API_PATH',
            default=('http://' + os.environ['LAB_HOST'] +':'
                + os.environ['LAB_PORT']), help='Path to the database.')
    parser.add_argument('-u',action='store',dest='USER',default='testuser',
            help='user name')
    parser.add_argument('-t',action='store',dest='DATASETS',
            help='turn on ai for these datasets')
    parser.add_argument('-n_recs',action='store',dest='N_RECS',type=int,
            default=1, help=('Number of recommendations to make at a time. '
                'If zero, will send continuous recommendations.'))
    parser.add_argument('-max_time',action='store',dest='MAX_TIME',type=int,
            default=60, help=('Amount of time to allow recs in seconds. '
                'Only works when term_condition set to "time".'))
    parser.add_argument('-term_condition',action='store',dest='TERM_COND',
            type=str, default='n_recs', choices=['n_recs','time','continuous'],
            help=('Termination condition for the AI.'))
    parser.add_argument('-v','-verbose',action='store_true',dest='VERBOSE',
            default=True, help='Print out more messages.')
    parser.add_argument('-warm',action='store_true',dest='WARM_START',
            default=False, help='Start from last saved session.')
    parser.add_argument('-sleep',action='store',dest='SLEEP_TIME',default=4,
            type=float, help='Time between pinging the server for updates')
    parser.add_argument('--knowledgebase','-k', action='store_true',
            dest='USE_KNOWLEDGEBASE', default=False,
            help='Load a knowledgebase for the recommender')

    args = parser.parse_args()

    # set logging level
    if args.VERBOSE:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    logger.debug(str(args))

    # dictionary of default recommenders to choose from at the command line.
    name_to_rec = {'random': RandomRecommender,
            'average': AverageRecommender,
            'knnmeta': KNNMetaRecommender,
            'svd': SVDRecommender,
            'cocluster': CoClusteringRecommender,
            'knnmeans': KNNWithMeansRecommender,
            'knndata': KNNDatasetRecommender,
            'knnml': KNNMLRecommender,
            'slopeone': SlopeOneRecommender
            }

    print('=======','Penn AI','=======')#,sep='\n')

    pennai = AI(rec_class=name_to_rec[args.REC],
            api_path=args.API_PATH, user=args.USER,
            verbose=args.VERBOSE, n_recs=args.N_RECS, 
            warm_start=args.WARM_START, datasets=args.DATASETS, 
            use_knowledgebase=args.USE_KNOWLEDGEBASE,
            term_condition=args.TERM_COND, max_time=args.MAX_TIME)

    n = 0;
    try:
        while True:
            # check for new experiment results
            if pennai.check_results():
                pennai.update_recommender()

            # check for user updates to request states
            pennai.check_requests()

            # process any active requests
            pennai.process_rec()

            n = n + 1
            time.sleep(args.SLEEP_TIME)

    except (KeyboardInterrupt, SystemExit):
        logger.info('Exit command recieved')
    except:
        logger.error("Unhanded exception caught: "+ str(sys.exc_info()[0]))
        raise
    finally:
        # shut down gracefully
        logger.info("Shutting down AI engine...")
        logger.info("...Shutting down Request Manager...")
        pennai.requestManager.shutdown()
        logger.info("Goodbye")

if __name__ == '__main__':
    main()



# Handles requests for the AI system.
import time
import ai.q_utils as q_utils
import logging
from enum import Enum, auto, unique
import ai.q_utils as q_utils

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

ch = logging.StreamHandler()
formatter = logging.Formatter('%(module)s: %(levelname)s: %(message)s')
ch.setFormatter(formatter)
logger.addHandler(ch)



@unique
class TerminalCondition(Enum):
    NUM_RECS = auto()   # generate a static number of recomendations
    TIME = auto()       # generate reccomendations for a set period of time
    CONTINUOUS = auto() # generate recomencations until the ai recommender is turned off by the user


class RequestManager:
    def __init__(self,
                ai,
                defaultTermConditionStr,
                defaultTermParam):
        self.ai = ai
        self.defaultTermParam = defaultTermParam

        if (defaultTermConditionStr == 'n_recs'):
            self.defaultTermCondition = TerminalCondition.NUM_RECS
        elif (defaultTermConditionStr == 'time'):
            self.defaultTermCondition = TerminalCondition.TIME
        elif (defaultTermConditionStr == 'continuous'):
            self.defaultTermCondition = TerminalCondition.CONTINUOUS
        else:
            msg = 'Unable to start RequestManager, unknown terminal condition: "' + str(defaultTermConditionStr) + '"'
            logger.error(msg)
            raise RuntimeError(msg)

        self.aiRequests = {} # dict of datasetId:aiRequest

        logger.info("initilized RequestManager.  defaultTermCondition:{}, defaultTermParam:{}".format(self.defaultTermCondition, self.defaultTermParam))


    def add_request(self, datasetId, datasetName):
        """ Add a new ai request for the given datasetId

        :param datasetId:
        """
        if(datasetId in self.aiRequests):
            req = self.aiRequests[datasetId]
        else:
            req = AiRequest(
                ai=self.ai,
                datasetId=datasetId,
                datasetName=datasetName)
            self.aiRequests[datasetId] = req


        req.new_request(
            termCondition=self.defaultTermCondition,
            termParam=self.defaultTermParam)


    def terminate_request(self, datasetId):
        """ Terminate requests for the given dataset id

        :param datasetId:
        """
        if (datasetId in self.aiRequests):
            req = self.aiRequests[datasetId]
            req.terminate_request()
        else:
            msg = 'Tried to terminate a dataset ai request before it had been initilized.  DatasetId: "' + str(datasetId) + '"'
            logger.info(msg)
            self.ai.labApi.set_ai_status(datasetId, 'finished')
        

    def process_requests(self):
        """ Process all active requests
        """
        for dataId, req in self.aiRequests.items():
            req.process_request()


    def shutdown(self):
        """ Terminate all requests and release all request threads
        """
        
        # release all datasetThreads
        q_utils.exitFlag = 1
        
        for dataId, req in self.aiRequests.items(): 
            req.terminate_request()



@unique
class AiState(Enum):
    INITILIZE = auto()
    WAIT_FOR_QUEUE_EMPTY = auto()
    ADD_RECOMMENDATIONS = auto()
    TERMINATE = auto()
    INACTIVE = auto()
    # python threads cannot be restarted; 
    THREAD_DIED_TERMINATE = auto() # uncrecoverable thread error detected; empty the queue, -> THREAD_DIED
    THREAD_DIED = auto() # unrecoverable thread error has happened


class AiRequest:
    def __init__(self, 
                ai, 
                datasetId, 
                datasetName):

        self.ai = ai
        self.datasetId = datasetId
        self.datasetName = datasetName
        self.datasetThread = q_utils.startQ(
            ai=self.ai, 
            datasetId=datasetId, 
            datasetName=datasetName)

        self.defaultRecBatchSize = 5
        self.state = AiState.INACTIVE

        logger.info("AiRequest initilized ({},{})".format(self.datasetName, self.datasetId))


    def new_request(self, termCondition, termParam):
        logger.info("AiRequest new_request ({},{})".format(self.datasetName, self.datasetId))
        self.termCondition = termCondition
        self.termParam = termParam
        self.startTime = time.time()

        if termCondition == TerminalCondition.NUM_RECS:
            self.recBatchSize = self.termParam
        else:
            self.recBatchSize = self.defaultRecBatchSize

        logger.info(f'new_request() termCondition={self.termCondition}; recBatchSize={self.recBatchSize}')

        self.state = AiState.INITILIZE


    def terminate_request(self, setServerAiState = True, nextState = AiState.INACTIVE):
        # get rid of everything in the queue
        # set state to inactive
        ##logger.debug("=======")
        ##logger.debug("=======")
        ##logger.debug("=======")
        logger.info(f'AiRequest terminate_request ({self.datasetName},{self.datasetId})')
        logger.debug(f'queue size: {self.datasetThread.workQueue.qsize()}')
        
        q_utils.removeAllExperimentsFromQueue(ai=self.ai,
                                    datasetId=self.datasetId)

        logger.debug(f'Removed experiments from queue, isQueueEmpty()={q_utils.isQueueEmpty(self.ai, self.datasetId)}')
        logger.debug(f'queue size: {self.datasetThread.workQueue.qsize()}')

        if (setServerAiState):
            self.ai.labApi.set_ai_status(self.datasetId, 'finished')

        self.state = nextState


    def process_request(self):
        logger.debug(f"===={self.datasetName} AI.State={self.state}, "
                f"queue={self.datasetThread.workQueue.qsize()},"
                f" processingRequest={self.datasetThread._processingRequest}====")
        logger.debug(f"====     _killActiveRequest={self.datasetThread._killActiveRequest}"
                f"  self.datasetThread.is_alive()={self.datasetThread.is_alive()}")
        if (self.state in [AiState.INACTIVE, AiState.THREAD_DIED]):
            return

        # update state as per termination strategy
        self.update_state()

        if (self.state == AiState.TERMINATE):
            self.terminate_request(setServerAiState=True)

        elif (self.state == AiState.THREAD_DIED_TERMINATE):
            logger.error(f'Queue error encountered; terminating request for {self.datasetName}')
            self.terminate_request(setServerAiState=True, nextState=AiState.THREAD_DIED)

        elif (self.state == AiState.WAIT_FOR_QUEUE_EMPTY):
            return

        elif (self.state == AiState.ADD_RECOMMENDATIONS):
            logger.debug("AiRequest adding recs ({},{},{})".format(self.datasetName, self.datasetId, self.recBatchSize))
            recs = self.ai.generate_recommendations(self.datasetId, 
                        self.recBatchSize)

            logger.info(f"recs: {recs}")

            logger.debug(f'queue size: {self.datasetThread.workQueue.qsize()}')
            q_utils.addExperimentsToQueue(ai=self.ai, 
                                     datasetId=self.datasetId, 
                                     experimentPayloads=recs)
            logger.debug(f'queue size: {self.datasetThread.workQueue.qsize()}')
            self.state = AiState.WAIT_FOR_QUEUE_EMPTY


    def update_state(self):
        if self.state == AiState.INACTIVE:
            msg = 'Tried to run update_state() when state was INACTIVE.  DatasetId: "' + str(datasetId) + '"  DatasetName: "' + str(datasetName) + '"'
            logger.error(msg)
            raise RuntimeError(msg)

        # if there was an uncaught error that killed the queue runner, terminate
        if not(self.datasetThread.is_alive()):
            self.state = AiState.THREAD_DIED_TERMINATE
            return

        # always start by adding recommendations
        if self.state == AiState.INITILIZE:
            q_utils.initilizeQueue(ai=self.ai, 
                            datasetId=self.datasetId)
            self.state = AiState.ADD_RECOMMENDATIONS
            return


        ###########
        # terminal conditions that only run set number of experiments
        ###########
        if self.termCondition == TerminalCondition.NUM_RECS:
            if (q_utils.isQueueEmpty(self.ai, self.datasetId) == True):
                logger.debug("AiRequest NUM_RECS terminal cond reached ({},{})".format(self.datasetName, self.datasetId))
                self.state=AiState.TERMINATE
                return
            else:
                return

        ###########
        # terminal conditions that continue until a terminal condition met
        #
        #   if the terminal condition is met, exit immediately.
        #
        #   if the terminal condition is not met and the queue is
        #   empty, add experiments.
        #
        #   otherwise, wait and let the queue continue processing
        ###########
        if (self.termCondition == TerminalCondition.TIME and
                time.time() - self.startTime > self.termParam):
            logger.debug("AiRequest TIME terminal cond reached ({},{})".format(self.datasetName, self.datasetId))
            self.state=AiState.TERMINATE
            return

        # add experiments if the queue is empty
        if (q_utils.isQueueEmpty(self.ai, self.datasetId) == True):
            logger.debug("AiRequest refilling queue ({},{})".format(self.datasetName, self.datasetId))
            self.state=AiState.ADD_RECOMMENDATIONS
            return



"""
Utility functions for loading experiment results that can be used for building
an initial knowledgebase for the recommenders.


A knowledgebase consists of a relational dataset of previously run experiments,
and the metafeatures of the datasets in those experiments.
"""

import pandas as pd
import numpy as np
import os
import ai.metalearning.get_metafeatures as mf
import logging
import json
import csv

# add trace level
logging.TRACE = logging.DEBUG + 5
logging.addLevelName(logging.TRACE, "TRACE")
class TraceLogger(logging.getLoggerClass()):
    def trace(self, msg, *args, **kwargs):
        self.log(logging.TRACE, msg, *args, **kwargs)
logging.setLoggerClass(TraceLogger)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
ch = logging.StreamHandler()
formatter = logging.Formatter('%(module)s: %(levelname)s: %(message)s')
ch.setFormatter(formatter)
logger.addHandler(ch)


PMLB_KB_CLASSIFICATION_RESULTS_PATH = ('data/knowledgebases/'
        'sklearn-benchmark-data-knowledgebase-r6.tsv.gz')
PMLB_KB_CLASSIFICATION_METAFEATURES_PATH = ('data/knowledgebases/'
        'pmlb_classification_metafeatures.csv.gz')
PMLB_KB_REGRESSION_RESULTS_PATH = ('data/knowledgebases/'
        'pmlb_regression_results.pkl.gz')
PMLB_KB_REGRESSION_METAFEATURES_PATH = ('data/knowledgebases/'
        'pmlb_regression_metafeatures.csv.gz')

USER_KB_RESULTS_PATH = 'data/knowledgebases/user/results'
USER_KB_METAFEATURES_PATH = 'data/knowledgebases/user/metafeatures'

def load_knowledgebase(resultsFiles={}, metafeaturesFiles=[],
        jsonMetafeatureDirectory='', dedupe=True):
    """Load experiment results from from file and generate metadata for the
    experiment datasets.

    :param resultsFiles - list<(string,string)>
        a list of tuples that contain resultsFile, pred_type
        - list of experiment results in tsv form, can be compressed files.
    :param metafeaturesFiles - list<tuple<string,string>>
        - list of files that contain metafeatures for the experiment datasets
        in csv form, can be compressed files.
    :param jsonMetafeatureDirectory
        - root of a directory structure that contains .json files for
        metafeatures

    :returns dict {
        dict {
        pred_type
        resultsData: DataFrame with columns corresponding to:
                    '_id',
                'dataset',
                'algorithm',
                'parameters',
                    classification or regression accuracy metrics
                    (i.e. 'accuracy', 'macrof1', 'bal_accuracy')

        metafeaturesData: Dataframe with columns corresponding to:
                    '_id',
                    '_metafeature_version',
                    '_prediction_type',
                    metafeature columns as defined in get_metafeatures.py

        warnings: list of warning Strings
                }
    """
    logger.info(f"load_knowledgebase('{resultsFiles}', {metafeaturesFiles}', "
            f"'{jsonMetafeatureDirectory}')")

    # load experiment results
    frames = []
    for resultsFile in resultsFiles:
        frames.append(_load_results_from_file(resultsFile))

    logger.info("concatenating results....")
    all_resultsData = pd.concat(frames, sort=False)

    if dedupe:
        logger.info("deduplicating results....")
        dedupe_results_dataframe(all_resultsData)
    dataset_names = all_resultsData['dataset'].unique()

    # load dataset metafeatures
    logger.info('load metafeatures...')
    metafeaturesDict = {}
    if jsonMetafeatureDirectory:
        metafeaturesDict.update(
                _load_json_metafeatures_from_directory(
                    jsonMetafeatureDirectory, dataset_names)
                )

    if metafeaturesFiles:
        assert isinstance(metafeaturesFiles, list), \
                (f"load_knowledgebase.metafeaturesFiles must be a list;"
                " got '{metafeaturesFiles}'")
        for mfFile in metafeaturesFiles:
            metafeaturesDict.update(_load_metadata_from_file(mfFile))

    if not(jsonMetafeatureDirectory or metafeaturesFiles):
        raise ValueError('One of metafeaturesFile or jsonMetafeatureDirectory '
                             'has to be specified')

    metafeaturesData = pd.DataFrame.from_records(
                metafeaturesDict).transpose()
    # filter any extraneous metafeaturesData (e.g. from extra datasets)
    metafeaturesData = metafeaturesData.loc[metafeaturesData['_id'].isin(
        all_resultsData['_id'].unique())]

    # check that all result datasets have metadata
    warnings = _validate_knowledgebase(all_resultsData, metafeaturesData)

    if (warnings):
        logger.warn(f"Warnings while running load_knowledgebase()"
            f"\n  resultsFiles: '{resultsFiles}'"
            f"\n  metafeaturesFiles: '{metafeaturesFiles}'"
            f"\n  jsonMetafeatureDirectory: '{jsonMetafeatureDirectory}'")

        logger.warn(f"Found {len(warnings)} warning(s):")
        logger.warn("\n".join(warnings))

    # split knowledgebases by prediction type
    resultsData = {}
    for pred_type in metafeaturesData['_prediction_type'].unique():
        relevant_dataset_ids = metafeaturesData.loc[
                metafeaturesData['_prediction_type']==pred_type,'_id']
        resultsData[pred_type] = all_resultsData.loc[
                all_resultsData['_id'].isin(relevant_dataset_ids)]
        # drop columns irrelevant to this pred_type
        resultsData[pred_type] = resultsData[pred_type].dropna(
                axis=1,how='all')

    return {'resultsData': resultsData, 'metafeaturesData': metafeaturesData,
            'warnings': warnings}


def dedupe_results_dataframe(resultsData):
    logger.trace("dedupe_results_dataframe()")

    logger.trace("generating parmHash")
    resultsData['parmHash'] = resultsData.apply(
            lambda x: hash(frozenset(x['parameters'].items())), axis = 1)
    logger.trace("parmHash generated")

    rawCount = len(resultsData)
    pd.set_option('display.max_columns', None)
    #logger.debug(resultsData.head())

    logger.trace("removing parameters....")
    compCols = list(resultsData.columns)
    compCols.remove("parameters")

    logger.trace("drop_duplicates...")
    resultsData.drop_duplicates(subset=compCols, inplace=True)

    logger.trace("drop paramHash")
    resultsData.drop(columns=['parmHash'], inplace=True)

    logger.info(f"loaded {rawCount} experiments, {len(resultsData)} "
            "after dropDuplicates")

    return resultsData


def load_default_knowledgebases(usePmlb=True,
        userKbResultsPath=USER_KB_RESULTS_PATH,
        userKbMetafeaturesPath=USER_KB_METAFEATURES_PATH,
        dedupe=True):
    """
    Convienence method to load the pmlb knowledgebase and any user-added
    knowledgebases
    """
    logger.info(f"load_default_knowledgebases('{usePmlb}', "
            f"'{userKbResultsPath}', '{userKbMetafeaturesPath}'")

    resFileExtensions = ['.json', 'pkl', '.tsv', '.gz']
    mfFileExtensions = ['.csv', '.tsv', '.gz']

    resultsFiles = []
    metafeaturesFiles = []

    # load pmlb
    if usePmlb:
        resultsFiles.append(
                PMLB_KB_CLASSIFICATION_RESULTS_PATH)
        resultsFiles.append(
                PMLB_KB_REGRESSION_RESULTS_PATH)
        metafeaturesFiles.append(
                PMLB_KB_CLASSIFICATION_METAFEATURES_PATH)
        metafeaturesFiles.append(
                PMLB_KB_REGRESSION_METAFEATURES_PATH)

    # if additional directories for results or metafeatures provided, use them
    if (userKbResultsPath):
        for root, dirs, files in os.walk(userKbResultsPath):
            for name in files:
                extension = os.path.splitext(name)[1]
                if (not name.startswith('.')
                        and (extension in resFileExtensions)):
                    resultsFiles.append(os.path.join(root, name))

    if (userKbMetafeaturesPath):
        for root, dirs, files in os.walk(userKbMetafeaturesPath):
            for name in files:
                extension = os.path.splitext(name)[1]
                if (not name.startswith('.')
                        and (extension in mfFileExtensions)):
                    metafeaturesFiles.append(os.path.join(root, name))

    return load_knowledgebase(
            resultsFiles = resultsFiles,
            metafeaturesFiles = metafeaturesFiles,
            dedupe=dedupe
            )

def generate_metafeatures_file(
    datasetDirectory,
    outputPath,
    outputFilename = "metafeatures.csv",
    predictionType = "classification",
    targetField = 'class',
    checkSubdirectories = True,
    fileExtensions = ['.csv', '.tsv'],
    **kwargs):
    """
    Generate metafeatures file for all datsets in the given directory

    """
    logger.info(f"generate_metafeatures_file({outputPath}, {datasetDirectory},"
            f" {outputFilename}, ...)")

    os.makedirs(outputPath, exist_ok=True)

    metafeaturesData = _generate_metadata_from_directory(
        datasetDirectory, predictionType, targetField,
        checkSubdirectories, fileExtensions, **kwargs)

    df = pd.DataFrame(metafeaturesData).transpose()

    logger.debug(f"generated metafeatures for {len(metafeaturesData)} "
            "datasets")
    #logger.debug(f"metafeaturesData: {metafeaturesData}")
    logger.debug(df.head())

    #, quoting=csv.QUOTE_NONNUMERIC)
    df.to_csv(os.path.join(outputPath, outputFilename), header=True,
            compression='gzip')

    return metafeaturesData


def _validate_knowledgebase(resultsDf, metafeaturesDf):
    """
    Validate knowledgebase
    """
    requiredResultsFields = ['_id', 'dataset', 'algorithm']
    requiredMetafeatureFields = ['_id', '_metafeature_version',
            '_prediction_type']

    warnings = []

    assert isinstance(resultsDf, pd.DataFrame)
    assert isinstance(metafeaturesDf, pd.DataFrame)

    # check that resultsDf has required fields
    for reqField in requiredResultsFields:
        if not reqField in resultsDf.columns:
            warnings.append("Knowledgebase experiments data "
                f"missing required field '{reqField}'")

    # check that metafeaturesDf has required fields
    for reqField in requiredMetafeatureFields:
        if not reqField in metafeaturesDf.columns:
            warnings.append("Knowledgebase metafeature data "
                f"missing required field '{reqField}'")

    if warnings:
        return warnings

    resMissingMf = resultsDf[~resultsDf['_id'].isin(metafeaturesDf['_id'])]
    resMissingMfUnique = resMissingMf[['_id', 'dataset']].drop_duplicates()

    if len(resMissingMfUnique.index) > 0:
        warnings.append(f"Found {len(resMissingMfUnique.index)} of "
              f"{len(resultsDf._id.unique())} experiment datasets with no "
              "associated metadata: " +  resMissingMfUnique.to_string()
              +'.\nResults missing metadata will not be loaded.')


    # check that all the metafeatures were created with a version compatable
    # with the current version of datasest_describe.py
    mfVersions = metafeaturesDf['_metafeature_version'].unique()
    if len(mfVersions) != 1:
        warnings.append(f'Multiple metafeature versions present: {mfVersions}')


    return warnings


def _load_results_from_file(resultsFile):
    """
    Load experiment results from file
    """
    logger.info(f'_load_results_from_file({resultsFile})')

    # resultsfile is a tsv file
    if resultsFile.endswith('tsv.gz') or resultsFile.endswith('tsv'):
        results_data = pd.read_csv(resultsFile, sep='\t')

        # convert params to dictionary
        logger.trace("converting parameters to dictionary")
        results_data['parameters'] = results_data['parameters'].apply(
                lambda x: eval(x))
        logger.info(f'returning {len(results_data)} results from {resultsFile}')
    elif resultsFile.endswith('.pkl') or resultsFile.endswith('.pkl.gz'):
        results_data = pd.read_pickle(resultsFile)
    elif resultsFile.endswith('.json') or resultsFile.endswith('.json.gz'):
        results_data = pd.read_json(resultsFile)
    else:
        raise ValueError("Unknown knowlegde base format!")
    assert not results_data.isna().any().any()
    logger.debug(f'results_data:\n{results_data.head()}')
    logger.trace("saving results_data to json "
                "from _load_results_from_file()")
    logger.trace("returning from _load_results_from_file()")
    return results_data


def _load_json_metafeatures_from_directory(metafeatureDirectory, datasetNames):
    """Load .json metafeatures for datasets

    Assumes metafeature files are named 'metafeatures.json' and are in
    subdirectories for each datset:
        metafeatureDirectory/dataset1/metadata.json
        metafeatureDirectory/dataset2/metadata.json
        metafeatureDirectory/dataset3/metadata.json

    :param metafeatureDirectory
    :param datasetNames - list of String dataset names
    """
    logger.info(f"Loading json metafeatures from directory "
            f"'{metafeatureDirectory}'")

    metafeaturesData = {}

    for dataset in np.unique(datasetNames):
        mfPath = os.path.join(metafeatureDirectory, dataset,
                'metafeatures.json')
        if os.path.exists(mfPath):
            #logger.debug(f'loading {mfPath}')
            with open(mfPath) as data_file:
                data = json.load(data_file)
            metafeaturesData[dataset] = data
        else:
            logger.warn(f"Couldn't find metafeature file for dataset "
                    f"'{dataset}'")

    return metafeaturesData

def _load_metadata_from_file(metafeaturesFile):
    logger.info(f"Loading metadata from file '{metafeaturesFile}")
    metafeaturesDf = pd.read_csv(metafeaturesFile, index_col=0,
            float_precision='round_trip') #, quoting=csv.QUOTE_NONNUMERIC)
    logger.debug("loaded metafeature file as df:")
    logger.debug(metafeaturesDf.head())

    return metafeaturesDf.to_dict(orient='index')

def _generate_metadata_from_directory(datasetDirectory,
    prediction_type = "classification", targetField = 'class',
    checkSubdirectories = True, fileExtensions = ['.csv', '.tsv'],
    **kwargs):
    """Extract metafeatures for all dataset files in the directory

    :returns dict: dataset name(str):metafeatures(dataFrame)
    """
    logger.info(f"generating metafeatures for files in directory "
            f"'{datasetDirectory}', targetField={targetField}, "
            f"checkSubdirectories={checkSubdirectories}, "
            f"fileExtensions={fileExtensions}")

    if (not datasetDirectory):
        raise ValueError("Could not generate metadata from directory, "
                "'datasetDirectory' must be specified")

    metafeaturesData = {}

    for root, dirs, files in os.walk(datasetDirectory):
        for name in files:
            print(name)
            extension = os.path.splitext(name)[1]
            # print('extension',extension, 'fileExtensions:',fileExtensions)
            if not name.startswith('.') and (extension in fileExtensions):
                # split twice to handle double extensions, i.e.
                #   'myfile.tsv.gz' => 'myfile'
                dataset = os.path.splitext(os.path.splitext(name)[0])[0]
                datapath = os.path.join(root, name)
                logger.info(f"Generating metadata for {datapath}")
                metafeatures = mf.generate_metafeatures_from_filepath(
                        datapath, prediction_type, targetField, **kwargs)
                metafeaturesData[dataset] = metafeatures

    return metafeaturesData



import pandas as pd
import dask.dataframe as dd
import json
import pymongo
from bson.objectid import ObjectId
import string
import urllib.request, urllib.parse
import json
from datetime import datetime
from tqdm import tqdm
import os
pd.options.mode.chained_assignment = None
#limit number of records we'll process
#
baseURL=os.environ['FGLAB_URL']
MONGODB_URI=os.environ['MONGODB_URI']
datasetsURL=baseURL + '/api/datasets'
algorithmsURL=baseURL + '/api/projects'
datasetsURL=baseURL + '/api/datasets'
apiDict = {'apikey':os.environ['APIKEY']}
apiParams = json.dumps(
    apiDict
    ).encode('utf8')
#
# first, we need to construct a dataset
print('loading api datasets into dictionary...')
req = urllib.request.Request(datasetsURL, data=apiParams,
  headers={'content-type': 'application/json'})
r = urllib.request.urlopen(req)
datasets = json.loads(r.read().decode(r.info().get_param('charset') or 'utf-8'))
datasets_dict = {}
for dataset in datasets:
     name =  dataset['name'].lower();
     _id = dataset['_id'];
     files = dataset['files'];
     datasets_dict[name] =  {'_id':_id,'files':files};
#
print('loading api algorithms into dictionary...')
req = urllib.request.Request(algorithmsURL, data=apiParams,
  headers={'content-type': 'application/json'})
r = urllib.request.urlopen(req)
algorithms = json.loads(r.read().decode(r.info().get_param('charset') or 'utf-8'))
algorithm_names = {}
for algorithm in algorithms:
     name =  algorithm['name'].lower();
     _id = algorithm['_id'];
     algorithm_names[name.lower()] =  _id;
#
print('loading pmlb results data...')
data = pd.read_csv('metalearning/sklearn-benchmark5-data-short.tsv.gz', sep='\t',
                   names=['dataset',
                         'classifier',
                         'parameters',
                         'accuracy',
                         'macrof1',
                         'bal_accuracy']).fillna('')


print('discarding duplicate dataset/classifier combinations...')
data = data.groupby(['dataset','classifier']).head(1).reset_index(drop=True)
print('formatting records for import...')
records = json.loads(data.T.to_json()).values()
ret_records = []
client = pymongo.MongoClient(MONGODB_URI)
db = client.FGLab
for record in records:
#split parameters into individual fields
  # 
  algorithm_name = record['classifier'].lower()
  if(algorithm_name in algorithm_names):
    new_record = {}
    new_record['_project_id'] = ObjectId(algorithm_names[algorithm_name])
    #dataset info
    dataset_name = record['dataset'].lower();
    if(dataset_name in datasets_dict):
      dataset = datasets_dict[dataset_name]
      new_record['_dataset_id'] = ObjectId(dataset['_id'])
      files = datasets_dict[dataset_name.lower()]['files']
      new_files = []
      for filedata in files:
        filedata['_id'] = ObjectId(filedata['_id'])
        new_files.append(filedata);
      new_record['files'] = new_files 
   
      
    #algorithm parameters
    parameters = record['parameters'].split(",");
    nested_parameters = {}
    for parameter in parameters:
      psplit = parameter.split("=")
      if(len(psplit) > 1):
        nested_parameters[psplit[0]] = psplit[1];
        new_record['_options'] = nested_parameters;
    #scores
    scores = {}
    scores['accuracy_score'] = record['accuracy']
    scores['balanced_accuracy'] = record['bal_accuracy']
    scores['f1_score'] = record['macrof1']
    new_record['_scores'] = scores;
    #
    new_record['_started'] = datetime.min
    new_record['_finished'] = datetime.min
    new_record['_status'] = 'success'
    new_record['username'] = 'pmlb'
    db.experiments.insert(new_record)
    
#print(ret_records);
#
#for record_dict in ret_records:
#  post_params = json.dumps(
#    {**record_dict, **apiDict}
#    ).encode('utf8')
#  print(post_params);
  



# Recommender system for Penn AI.
import pandas as pd
# import json
# import urllib.request, urllib.parse
from .base import BaseRecommender
#from ..metalearning import get_metafeatures
from sklearn.preprocessing import RobustScaler
from sklearn.pipeline import Pipeline
import numpy as np
from collections import defaultdict, OrderedDict
from sklearn.neighbors import NearestNeighbors
import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
ch = logging.StreamHandler()
formatter = logging.Formatter('%(module)s: %(levelname)s: %(message)s')
ch.setFormatter(formatter)
logger.addHandler(ch)

class KNNMetaRecommender(BaseRecommender):
    """Penn AI KNN meta recommender.
    Recommends machine learning algorithms and parameters as follows:
        - store the best ML + P on every dataset.
        - given a new dataset, measure its distance to all results in
        metafeature space.
        - recommend ML + P with best performance on closest dataset.

    Parameters
    ----------
    ml_type: str, 'classifier' or 'regressor'
        Recommending classifiers or regressors. Used to determine ML options.
    metric: str (default: accuracy for classifiers, mse for regressors)
        The metric by which to assess performance on the datasets.
    ml_p: DataFrame (default: None)
        Contains all valid ML parameter combos, with columns 'algorithm' and
        'parameters'
    """
    def __init__(self,
            ml_type='classifier',
            metric=None,
            ml_p=None,
            random_state=None,
            knowledgebase_results=None,
            knowledgebase_metafeatures=None,
            load_serialized_rec="if_exists",
            serialized_rec_directory=None,
            serialized_rec_filename=None):

        """ set default recommender specific parameters; might be overwritten by loading serialized recommender"""

        # lookup table: dataset name to best ML+P
        self.best_mlp = pd.DataFrame(columns=['_id','algorithm',
            'parameters', 'score'])
        self.best_mlp.set_index('_id',inplace=True)

        # local dataframe of datasets and their metafeatures
        self.all_dataset_mf = pd.DataFrame()


        """Initialize recommendation system."""
        super().__init__(
            ml_type,
            metric,
            ml_p,
            random_state=random_state,
            knowledgebase_results=knowledgebase_results,
            knowledgebase_metafeatures=knowledgebase_metafeatures,
            load_serialized_rec=load_serialized_rec,
            serialized_rec_directory=serialized_rec_directory,
            serialized_rec_filename=serialized_rec_filename)


    def update(self, results_data, results_mf, source='pennai'):
        """Update ML / Parameter recommendations.

        Parameters
        ----------
        results_data: DataFrame
                columns corresponding to:
                'algorithm'
                'parameters'
                self.metric

        results_mf: DataFrame
               columns corresponding to metafeatures of each dataset in
               results_data.
        """
        # update trained dataset models and hash table
        super().update(results_data, results_mf, source)

        # save a copy of the results_mf with NaNs filled with zero
        drop_cols = [c for c in results_mf.columns
                if c[0] == '_' and c !='_id']
        self.all_dataset_mf = \
        results_mf.drop(columns=drop_cols).fillna(0.0).set_index('_id')

        # update internal model
        self.update_model(results_data)

    def update_model(self,results_data):
        """Stores best ML-P on each dataset."""
        logger.debug('len(self.hash_2_param)): ' + str(len(self.hash_2_param)))
        for d,dfg in results_data.groupby('_id'):
            if (len(self.best_mlp) == 0 or
                d not in self.best_mlp.index or
                dfg[self.metric].max() > self.best_mlp.loc[d,'score']):
                self.best_mlp.loc[d,'score'] = dfg[self.metric].max()
                dfg = dfg.reset_index()
                idx = dfg[self.metric].idxmax()
                # print('dfg:\n',dfg)
                logger.debug('new best for '+d+': '+
                        dfg.loc[idx,'algorithm']+', idx:'+str(idx))
                self.best_mlp.loc[d,'algorithm'] = dfg.loc[idx,'algorithm']
                self.best_mlp.loc[d,'parameters'] = dfg.loc[idx,
                        'parameter_hash']
            else:
                logger.debug('skipping'+d)
        # print('model updated')

    def recommend(self, dataset_id, n_recs=1, dataset_mf = None):
        """Return a model and parameter values expected to do best on dataset.

        Parameters
        ----------
        dataset_id: string
            ID of the dataset for which the recommender is generating
            recommendations.
        n_recs: int (default: 1), optional
            Return a list of length n_recs in order of estimators and parameters
            expected to do best.
        """
        if dataset_mf is None:
            raise ValueError('dataset_mf is None for',dataset_id,
                    "can't recommend")

        # dataset hash table
        super().recommend(dataset_id, n_recs, dataset_mf)

        logger.debug('dataset_mf columns:{}'.format(dataset_mf.columns))
        drop_cols = [c for c in dataset_mf.columns
                if c[0] == '_' and c !='_id']
        dataset_mf = dataset_mf.drop(columns=drop_cols)
        logger.debug('dataset_mf columns:{}'.format(dataset_mf.columns))
        try:
            ml_rec, phash_rec, rec_score = self.best_model_prediction(
                    dataset_id, dataset_mf)
            if len(ml_rec) < n_recs:
                logger.info(f'len(ml_rec)={len(ml_rec)}, recommending random')
            iters = 0
            while len(ml_rec) < n_recs and iters < 1000:
                # add random ml_p recommendations until n_recs is met
                new_ml_rec = np.random.choice(self.ml_p['algorithm'].unique())
                new_phash_rec = self._hash_simple_dict(np.random.choice(
                        self.ml_p.loc[self.ml_p['algorithm']==new_ml_rec]
                                              ['parameters'].values))
                if (dataset_id + '|' + new_ml_rec + '|' + new_phash_rec
                        not in self.trained_dataset_models):
                    ml_rec.append(new_ml_rec)
                    phash_rec.append(new_phash_rec)
                    rec_score.append(np.nan)
                iters = iters+1
            if iters == 1000:
                logger.info(f'couldn''t find {n_recs} unique recommendations! '
                      'returning',len(ml_rec))
                subset = \
                    [dataset_id in tdm for tdm in self.trained_dataset_models]
                num_results = len(
                    [tdm for i,tdm in enumerate(self.trained_dataset_models)
                           if subset[i]])
                logger.info(f'btw, there are {num_results} results for '
                        '{dataset_id} already')
            ml_rec, p_rec, rec_score = (ml_rec[:n_recs],
                    [self.hash_2_param[p] for p in phash_rec[:n_recs]],
                                       rec_score[:n_recs])
            assert(len(ml_rec) == n_recs)

        except Exception as e:
            logger.error('error running self.best_model_prediction for'
                    +dataset_id)
            raise e
            # logger.error('ml_rec:'+ ml_rec)
            # logger.error('p_rec'+ p_rec)
            # logger.error('rec_score'+rec_score)

        # update the recommender's memory with the new algorithm-parameter
        # combos that it recommended
        self._update_trained_dataset_models_from_rec(
                                                    dataset_id,
                                                    ml_rec,
                                                    phash_rec
                                                    )

        return ml_rec, p_rec, rec_score

    def best_model_prediction(self, dataset_id, df_mf, n_recs=1):
        """Predict scores over many variations of ML+P and pick the best"""
        # get dataset metafeatures
        for cols in ['dataset','_id']:
            if cols in df_mf.columns:
                df_mf = df_mf.drop(cols,axis=1)
        mf = df_mf.fillna(0.0).values.flatten()
        # compute the neighbors of past results
        nbrs = NearestNeighbors(n_neighbors=len(self.all_dataset_mf),
                algorithm='ball_tree')
        rs = RobustScaler()

        X = rs.fit_transform(self.all_dataset_mf.values)
        nbrs.fit(X)
        # find n_recs nearest neighbors to new dataset
        # print('querying neighbors with mf of shape',mf.shape)
        distances,indices = nbrs.kneighbors(rs.transform(mf.reshape(1,-1)))
        # print('distances:',distances)
        # print('indices:',indices)
        dataset_idx = [self.all_dataset_mf.index[i] for i in indices[0]]
        # recommend the mlp results closest to the dataset in metafeature space
        ml_recs, p_recs, scores = [],[],[]

        # print('self.best_mlp:',self.best_mlp)
        for i,(d,dist) in enumerate(zip(dataset_idx,distances[0])):
            if d not in self.best_mlp.index:
                continue
            if i < 10:
                logger.debug('closest dataset:'+d+'; distance:'+ str(dist))
            # don't recommend based on the same dataset
            if round(dist,6) > 0.0:
                alg_params = (self.best_mlp.loc[d,'algorithm'] + '|' +
                              self.best_mlp.loc[d,'parameters'])
                # only recommend if not already recommended
                if (dataset_id+'|'+alg_params
                        not in self.trained_dataset_models):
                    ml_recs.append(self.best_mlp.loc[d,'algorithm'])
                    p_recs.append(self.best_mlp.loc[d,'parameters'])
                    scores.append(dist)

        return ml_recs,p_recs,scores



import urllib.request, urllib.parse
import json
import os
baseURL = os.environ['FGLAB_URL']
experimentsURL=baseURL+'/api/datasets'
postvars = {'apikey':os.environ['APIKEY'],'ai':'requested'} 
params = json.dumps(postvars).encode('utf8')
req = urllib.request.Request(experimentsURL, data=params,
  headers={'content-type': 'application/json'})
r = urllib.request.urlopen(req)
data = json.loads(r.read().decode(r.info().get_param('charset') or 'utf-8'))
print(data)



"""
Recommender system for Penn AI.
"""
import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
ch = logging.StreamHandler()
formatter = logging.Formatter('%(module)s: %(levelname)s: %(message)s')
ch.setFormatter(formatter)
logger.addHandler(ch)
import numpy as np
import os
import pdb
import pickle
import gzip
import random
import hashlib
import copy
from pandas.util import hash_pandas_object
import pandas as pd

# implementing metaclass __repr__ for more human readable
# names for generated tests in test_recommender.py
class MC(type):
    def __repr__(self):
        return self.__qualname__

class BaseRecommender(object, metaclass=MC):
    """Base recommender for PennAI

    The BaseRecommender is not intended to be used directly; it is a skeleton class
    defining the interface for future recommenders within the PennAI project.

    Parameters
    ----------
    ml_type: str, 'classifier' or 'regressor'
        Recommending classifiers or regressors. Used to determine ML options.

    metric: str (default: accuracy for classifiers, mse for regressors)
        The metric by which to assess performance on the datasets.

    ml_p: DataFrame (default: None)
        Contains all valid ML parameter combos, with columns 'algorithm' and
        'parameters'

    knowledgebase_results: Pandas DataFrame or None
        Initial knowledgebase results data.
        If not None and not loading a serialized recommender, the recommender
        will initialize and train on this data.
        If loading a serialized recommender, this is the knowlegebase that
        accompanies it.

    knowledgebase_metafeatures: Pandas DataFrame or None
        Initial knowledgebase metafeatures data.
        If loading a serialized recommender, this is the knowlegebase that
        accompanies it.

    serialized_rec_directory: string or None
        Name of the directory to save/load a serialized recommender.
        Default directory is "."

    serialized_rec_filename: string or None
        Name of the file to save/load a serialized recommender.
        If the filename is not provided, the default filename based on the
        recommender type, and metric, and knowledgebase used.

    load_serialized_rec: str, "always", "never", "if_exists"
        Whether to attempt to load a serialized recommender:
            "if_exists" - If a serialized recomender exsists at the specified
            path, load it.
            "always" - Always load a serialized recommender.  Throw an
            exception if no serialized recommender exists.
            "never" - Never load a serialized recommender.

    """

    def __init__(self,
            ml_type='classifier',
            metric=None,
            ml_p=None,
            random_state=None,
            knowledgebase_results=None,
            knowledgebase_metafeatures=None,
            load_serialized_rec="if_exists",
            serialized_rec_directory=None,
            serialized_rec_filename=None):

        """Initialize recommendation system."""
        if ml_type not in ['classifier', 'regressor']:
            raise ValueError('ml_type must be "classifier" or "regressor"')

        if load_serialized_rec not in ["always", "never", "if_exists"]:
            raise ValueError('load_serialized_rec must be "always", "never" or'
                    ' "if_exists"')

        self.random_state = random_state
        if self.random_state is not None:
            random.seed(self.random_state)
            np.random.seed(self.random_state)

        logger.info('self.random_state: ' + str(self.random_state))

        self.ml_type = ml_type

        if metric is None:
            self.metric='bal_accuracy' if self.ml_type=='classifier' else 'mse'
        else:
            self.metric = metric

        # maintain a set of dataset-algorithm-parameter combinations that have
        # already been evaluated
        self.trained_dataset_models = set()
        # hash table for parameter options
        self.hash_2_param = {}

        # get ml+p combos (note: this triggers a property in base recommender)
        # self.ml_htable = {}
        self.ml_p = ml_p

        # generate the serialized recommender path
        self.serialized_rec_path = self._generate_serialized_rec_path(
            serialized_rec_filename,
            serialized_rec_directory
            )


        # train an empty recommender, either using the provided kb or
        # loading a serialized rec from file
        self._train_empty_rec(
            ml_type,
            metric,
            ml_p,
            random_state,
            knowledgebase_results,
            knowledgebase_metafeatures,
            load_serialized_rec,
            serialized_rec_directory,
            serialized_rec_filename)


    def _train_empty_rec(self,
            ml_type,
            metric,
            ml_p,
            random_state,
            knowledgebase_results,
            knowledgebase_metafeatures,
            load_serialized_rec,
            serialized_rec_directory,
            serialized_rec_filename):

        # load serialized rec, or initialize from the given
        # knowledgebase
        logger.info(f"load_serialized_rec='{load_serialized_rec}'")

        if load_serialized_rec == "always":
            if not os.path.exists(self.serialized_rec_path):
                raise ValueError(f"load_serialized_rec='{load_serialized_rec}'"
                        " but cannot load serialized recommender:"
                        " '{self.serialized_rec_path}'")
            self.load(self.serialized_rec_path, knowledgebase_results)

        elif load_serialized_rec == "if_exists":
            if os.path.exists(self.serialized_rec_path):
                logger.info(f"Loading serialized recommender:"
                        " {self.serialized_rec_path}")
                self.load(self.serialized_rec_path, knowledgebase_results)
            else:
                logger.warn(f"Not loading serialized recommender, file does "
                        "not exist: {self.serialized_rec_path}")
                if knowledgebase_results is not None:
                    logger.info(f"Initializing new recommender from provided "
                            "knowledgebase")
                    self.update(knowledgebase_results,
                            knowledgebase_metafeatures, source='knowledgebase')

        else:
            logger.info(f"Not loading serialized recommender.")
            if knowledgebase_results is not None:
                logger.info(f"Initializing new recommender from provided "
                        "knowledgebase")
                self.update(knowledgebase_results, knowledgebase_metafeatures,
                        source='knowledgebase')


    def _default_serialized_rec_filename(self):
        """Generate the default name of the serialized instance of this
        recommender
        """

        # Hardcoading the informal kb descriptor for now, this should be changed.
        return (
            self.__class__.__name__
            + '_' + self.ml_type
            + '_' + self.metric
            + '_pmlb_20200821'
            +'.pkl.gz')


    def _generate_serialized_rec_path(self,
        serialized_rec_filename=None,
        serialized_rec_directory=None):
        """ Generate the path to save/load serialized recommender

        Parameters
        ----------
        serialized_rec_filename
        serialized_rec_directory
        """

        # dynamic default values
        serialized_rec_directory = serialized_rec_directory or "."
        serialized_rec_filename = serialized_rec_filename or \
                self._default_serialized_rec_filename()

        return os.path.join(serialized_rec_directory, serialized_rec_filename)


    def update(self, results_data, results_mf=None, source='pennai'):
        """Update ML / Parameter recommendations.

        Parameters
        ----------
        results_data: DataFrame
            columns corresponding to:
            'algorithm'
            'parameters'
            self.metric

        results_mf: DataFrame, optional
            columns corresponding to metafeatures of each dataset in
            results_data.

        source: string
            if 'pennai', will update tally of trained dataset models
        """
        assert(results_data is not None), "results_data cannot be None"

        if results_data.isna().values.any():
            logger.warning('There are NaNs in results_data.')
            #logger.warning(str(results_data))
            logger.warning(results_data.head())
            logger.error('Dropping NaN results.')
            results_data.dropna(inplace=True)

        # update parameter hash table
        logger.info('updating hash_2_param...')
        self.hash_2_param.update(
                {self._hash_simple_dict(x):x
                for x in results_data['parameters'].values})
        param_2_hash = {frozenset(v.items()):k
                for k,v in self.hash_2_param.items()}
        # store parameter_hash variable in results_data
        logger.info('storing parameter hash...')
        results_data['parameter_hash'] = results_data['parameters'].apply(
                lambda x: param_2_hash[frozenset(x.items())])

        # update results list
        if source == 'pennai':
            self._update_trained_dataset_models_from_df(results_data)

    def _hash_simple_dict(self, x):
        """Provides sha256 hash for a dictionary with hashable items."""
        hasher = hashlib.sha256()
        hasher.update(repr(tuple(sorted(x.items()))).encode())
        return hasher.hexdigest()

    def recommend(self, dataset_id=None, n_recs=1, dataset_mf=None):
        """Return a model and parameter values expected to do best on dataset.

        Parameters
        ----------

        dataset_id: string
            ID of the dataset for which the recommender is generating
            recommendations.
        n_recs: int (default: 1), optional
            Return a list of length n_recs in order of estimators and parameters
            expected to do best.
        dataset_mf: DataFrame
            metafeatures of the dataset represented by dataset_id
        """
        # self.dataset_id_to_hash.update(
        #         {dataset_id:dataset_mf['_id'].values[0]})

    def load(self, filename=None, knowledgebase=None):
        """Load a saved recommender state.

        :param filename: string or None
            Name of file to load
        :param knowledgebase: string or None
            DataFrame with columns corresponding to:
                'dataset'
                'algorithm'
                'parameters'
                self.metric
        """
        if filename is None:
            fn = self.serialized_rec_path
        else:
            fn = filename

        if os.path.isfile(fn):
            logger.info('loading recommender ' + fn + ' from file')
            f = gzip.open(fn, 'rb')
            tmp_dict = pickle.load(f)
            f.close()

            #logger.debug(f"rec keys: {tmp_dict.keys()}")

            # check if parameters match, if not throw warning/error
            for k,v in tmp_dict.items():
                if k in self.__dict__.keys():
                    try:
                        if self.__dict__[k] != tmp_dict[k]:
                            logger.warn(k+' changing from '
                                    + str(self.__dict__[k])[:20] + '... to '
                                        + str(tmp_dict[k])[:20] + '...')
                    except:
                        pass
                else:
                    logger.warn('adding ' + k+'=' + str(tmp_dict[k])[:20]
                            + '...')
            logger.info('updating internal state')

            # check ml_p hashes
            rowHashes = hash_pandas_object(self.ml_p.apply(str)).values
            newHash = hashlib.sha256(rowHashes).hexdigest()
            if 'ml_p_hash' in tmp_dict.keys():
                if newHash == tmp_dict['ml_p_hash']:
                    logger.info('ml_p hashes match')
                else:
                    error_msg = ('the ml_p hash from the pickle is different. '
                        'This likely means the algorithm configurations have '
                        'changed since this recommender was saved. You should '
                        'update and save a new one.')
                    logger.error(error_msg)

                    # debugging
                    if ('_ml_p' in tmp_dict):
                        pd.testing.assert_frame_equal(self.ml_p, tmp_dict['_ml_p'])
                    else:
                        logger.error(f"Pickle does not contain _ml_p for debugging.")
                        logger.error(f"Keys: {tmp_dict.keys()}")

                    raise ValueError(error_msg)
                del tmp_dict['ml_p_hash']

            # update self with loaded pickle
            self.__dict__.update(tmp_dict)
            return True
        else:
            logger.warning('Could not load filename '+ fn)
            return False


    def save(self, filename=None):
        """Save the current recommender.

        :param filename: string or None
            Name of file to load
        """
        if filename is None:
            fn = self.serialized_rec_path
        else:
            fn = filename
        if os.path.isfile(fn):
            logger.warning('overwriting ' + fn)

        save_dict = copy.deepcopy(self.__dict__)

        # remove results_df to save space. this gets loaded by load() fn.
        if 'results_df' in save_dict.keys():
            logger.debug('deleting save_dict[results_df]:'
                    +str(save_dict['results_df'].head()))
            rowHashes = hash_pandas_object(save_dict['results_df']).values
            save_dict['results_df_hash'] = hashlib.sha256(
                    rowHashes).hexdigest()
            del save_dict['results_df']

        # remove ml_p to save space
        rowHashes = hash_pandas_object(save_dict['_ml_p'].apply(str)).values
        save_dict['ml_p_hash'] = hashlib.sha256(rowHashes).hexdigest()
        del save_dict['_ml_p']
        del save_dict['mlp_combos']

        logger.info('saving recommender as ' + fn)
        f = gzip.open(fn, 'wb')
        pickle.dump(save_dict, f, 2)
        f.close()

    def update_and_save(self, results_data, results_mf=None, source='pennai',
            filename=None):
        """runs self.update() and self.save.

        Parameters
        ----------
        results_data: DataFrame
            columns corresponding to:
            'algorithm'
            'parameters'
            self.metric

        results_mf: DataFrame, optional
            columns corresponding to metafeatures of each dataset in
            results_data.

        source: string
            if 'pennai', will update tally of trained dataset models
        """
        self.update(results_data, results_mf, source)
        self.save(filename)

    @property
    def ml_p(self):
        logger.debug('getting ml_p')
        return self._ml_p

    @ml_p.setter
    def ml_p(self, value):
        logger.debug('setting ml_p')
        if value is not None:
            #filter out SVC (temporary)
            self._ml_p = value[['algorithm','parameters']]
            logger.debug('setting hash table')
            # maintain a parameter hash table for parameter settings
            # if 'alg_name' not in value.columns:
            #     self._ml_p['alg_name'] = self._ml_p['algorithm']

            self.hash_2_param = {
                    self._hash_simple_dict(x):x
                    for x in self._ml_p['parameters'].values}
            param_2_hash = {frozenset(v.items()):k
                    for k,v in self.hash_2_param.items()}
            # machine learning - parameter combinations
            self.mlp_combos = (self._ml_p['algorithm']+'|'+
                               self._ml_p['parameters'].apply(lambda x:
                                   param_2_hash[frozenset(x.items())]))
            # filter out duplicates
            self.mlp_combos = self.mlp_combos.drop_duplicates()
            # # set ml_htable
            # if 'alg_name' in value.columns:
            #     self.ml_htable = {
            #             k:v for v,k in zip(value['alg_name'].unique(),
            #             value['algorithm'].unique())
            #             }
        else:
            logger.warning('value of ml_p is None')
        logger.debug('param_2_hash:{} objects'.format(len(param_2_hash)))

    def _update_trained_dataset_models_from_df(self, results_data):
        '''stores the trained_dataset_models to aid in filtering repeats.'''
        results_data.loc[:, 'dataset-algorithm-parameters'] = (
                                       results_data['_id'].values + '|' +
                                       results_data['algorithm'].values + '|' +
                                       results_data['parameter_hash'].values)

        for i,phash in enumerate(results_data['parameter_hash'].values):
            if phash not in self.hash_2_param.keys():
                logger.error(phash
                        +' not in self.hash_2_param. parameter values: '
                        + str(results_data['parameters'].values[i]))
        # get unique dataset / parameter / classifier combos in results_data
        d_ml_p = results_data['dataset-algorithm-parameters'].unique()
        self.trained_dataset_models.update(d_ml_p)

    def _update_trained_dataset_models_from_rec(self, dataset_id, ml_rec,
            phash_rec):
        '''update the recommender's memory with the new algorithm-parameter
        combos that it recommended'''
        if dataset_id is not None:
            # datahash = self.dataset_id_to_hash[dataset_id]
            self.trained_dataset_models.update(
                                    ['|'.join([dataset_id, ml, p])
                                    for ml, p in zip(ml_rec, phash_rec)])



# first, we need to construct a dataset that has each parameter as a feature. to do this we have to deconstruct
#the 'parameter' column into new columns.
import pandas as pd
import dask.dataframe as dd
import json
import pymongo
from bson.objectid import ObjectId
import string
import urllib.request, urllib.parse
import json
from datetime import datetime
from tqdm import tqdm
import os
pd.options.mode.chained_assignment = None
#limit number of records we'll process
max_records = 5000000000;
#
baseURL=os.environ['FGLab'] + '/api';
print(baseURL);
datasetsURL=baseURL + '/datasets'
algorithmsURL=baseURL + '/projects'
datasetsURL=baseURL + '/datasets'
apiDict = {'apikey':os.environ['apikey']}
apiParams = json.dumps(
    apiDict
    ).encode('utf8')
#
print('loading api datasets into dictionary...')
req = urllib.request.Request(datasetsURL, data=apiParams,
  headers={'content-type': 'application/json'})
r = urllib.request.urlopen(req)
datasets = json.loads(r.read().decode(r.info().get_param('charset') or 'utf-8'))
datasets_dict = {}
for dataset in datasets:
     name =  dataset['name'].lower();
     _id = dataset['_id'];
     files = dataset['files'];
     datasets_dict[name] =  {'_id':_id,'files':files};
#
print('loading api algorithms into dictionary...')
req = urllib.request.Request(algorithmsURL, data=apiParams,
  headers={'content-type': 'application/json'})
r = urllib.request.urlopen(req)
algorithms = json.loads(r.read().decode(r.info().get_param('charset') or 'utf-8'))
algorithm_names = {}
for algorithm in algorithms:
     name =  algorithm['name'].lower();
     _id = algorithm['_id'];
     algorithm_names[name.lower()] =  _id;
#
print('loading pmlb results data...')
#data = pd.read_csv('sklearn-benchmark5-data-edited.tsv.gz', sep='\t',
data = pd.read_csv('sklearn-benchmark5-data-short.tsv.gz', sep='\t',
                   names=['dataset',
                         'classifier',
                         'parameters',
                         'accuracy',
                         'macrof1',
                         'bal_accuracy']).fillna('')


print('formatting records for import...')
records = json.loads(data.T.to_json()).values()
ret_records = []
for record in records:
#split parameters into individual fields
  parameters= record['parameters'].split(",");
  nested_parameters = {}
  for parameter in parameters:
    psplit = parameter.split("=")
    if(len(psplit) > 1):
      nested_parameters[psplit[0]] = psplit[1];
# 
  algorithm_name = record['classifier']
  if(algorithm_name.lower() in algorithm_names):
    new_record = {}
    dataset_name = record['dataset']
    if(dataset_name.lower() in datasets_dict):
      dataset_id = ObjectId(datasets_dict[dataset_name.lower()]['_id'])
      dataset_files = datasets_dict[dataset_name.lower()]['files']
      new_record['_dataset_id'] = dataset_id;
      new_record['_files'] = dataset_files;
    #
    algorithm_id = ObjectId(algorithm_names[algorithm_name.lower()])
    new_record['_project_id'] = algorithm_id
    new_record['_options'] = nested_parameters;
    scores = {}
    scores['accuracy_score'] = record['accuracy']
    scores['balanced_accuracy'] = record['bal_accuracy']
    new_record['_scores'] = scores;
    new_record['_macrof1'] = record['macrof1']
    new_record['_started'] = datetime.now()
    new_record['_finished'] = datetime.now()
    new_record['_status'] = 'success'
    new_record['_files'] = []
    ret_records.append(new_record);
  if(len(ret_records) >= max_records):
    break
    
#print(ret_records);
#
client = pymongo.MongoClient()
db = client.FGLab
db.experiments.insert(ret_records)
#for record_dict in ret_records:
#  post_params = json.dumps(
#    {**record_dict, **apiDict}
#    ).encode('utf8')
#  print(post_params);
  



"""
Submission Queue functions for Penn-AI
"""

import queue
import threading
import time
import requests
import logging
import traceback
import sys

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

exitFlag = 0


class DatasetThread (threading.Thread):
    """Thread for managing pending ml experiments for a particular dataset"""
    def __init__(self, ai, datasetId, datasetName):
        threading.Thread.__init__(self)

        self.datasetId = datasetId # database id of the dataset
        self.name = datasetName # name of the dataset for the given datasetId      
        self.workQueue = queue.Queue() # queue of experiment payloads
        self.ai = ai # instance of an ai object

        self._prLock = threading.Lock()
        self._processingRequest = False # true if process_data() is actively processing a request
        
        # multi-thread variables
        self._killActiveRequest = False # true if there was a request to kill all active requests.  
        self._karLock = threading.Lock() #lock for thread-safe access to to access to _killActiveRequest

        # Handle thread exceptions 
        # see <https://stackoverflow.com/questions/2829329/catch-a-threads-
        # exception-in-the-caller-thread-in-python>
        self.__threadExceptionBucket = queue.Queue()

    def run(self):
        """This overrides threading.Thread's run method, 
        and is invoked by start()"""
        logger.debug(f"Creating queue for thread {self.name}")

        while not exitFlag:
            try:
                logger.info(f"Starting loop for thread {self.name}")
                process_data(self)
            except:
                exc_type, exc_obj, exc_tb = sys.exc_info()
                logger.error("=========================================================================")
                logger.error("=========================================================================")
                logger.error("=========================================================================")
                logger.error("=========================================================================")
                logger.error("=========================================================================")
                logger.error("Exception caught in thread '" + self.name + "': " + 
                             str(exc_obj))
                logger.error(''.join(traceback.format_exception(exc_type, exc_obj, 
                                                                exc_tb))) 
            #see <https://docs.python.org/3/library/traceback.html#traceback-examples>
                self.__threadExceptionBucket.put(exc_obj)


        logger.info(f"Exiting queue for thread {self.name}")


def startQ(ai, datasetId, datasetName):
    """Get or start a threaded queue for experiments for a 
    particular dataset.  Only one queue is created per dataset.

    :param ai: ai.AI - instance of AI class
    :param datasetId: string - datasetId for which to start a queue

    :returns: DatasetThread
    """
    if(datasetId in ai.dataset_threads):
        dsThread = ai.dataset_threads[datasetId]
    else:
        dsThread = DatasetThread(ai, datasetId, datasetName)
        dsThread.start()
    ai.dataset_threads[datasetId] = dsThread

    return dsThread
    
def addExperimentsToQueue(ai, datasetId, experimentPayloads):
    for payload in experimentPayloads:
        addExperimentToQueue(ai, datasetId, payload)

def addExperimentToQueue(ai, datasetId, experimentPayload):
    """Add ml experiment to the queue for a particular dataset

    :param ai: ai.AI - instance of AI class
    :param datasetId: string - datasetId to queue an experiment for
    :param experimentPayload: dict -  payload that describes the ml experiment
    """
    dsThread = ai.dataset_threads[datasetId]
    workQueue = dsThread.workQueue
    workQueue.put(experimentPayload)

def removeAllExperimentsFromQueue(ai, datasetId):
    dsThread = ai.dataset_threads[datasetId]    
    workQueue = dsThread.workQueue

    with dsThread._karLock:
        dsThread._killActiveRequest = True

        try:
            while True:
                workQueue.get_nowait()
        except queue.Empty:
            pass

def initilizeQueue(ai, datasetId):
    dsThread = ai.dataset_threads[datasetId]    
    workQueue = dsThread.workQueue

    with dsThread._karLock:
        dsThread._killActiveRequest = False


def isQueueEmpty(ai, datasetId):
    dataset_thread = ai.dataset_threads[datasetId]

    with dataset_thread._prLock:
        isEmpty = dataset_thread.workQueue.empty() and (dataset_thread._processingRequest == False)
    
    return isEmpty


def process_data(dsThread):
    """Loop to monitor and submit all experiments in the datasetThreads workQueue
    Exits only when exitFlag is true.

    Called by DatasetThread.run()  

    :param dsThread: DatasetThread
    """
    logger.debug("process_data("+ str(dsThread) + ") exitFlag:" + str(exitFlag))

    while not exitFlag:
        try:
            data = dsThread.workQueue.get(timeout=.1) #blocks until timeout

            try:
                with dsThread._prLock:
                    dsThread._processingRequest = True
                logger.debug(f"==={dsThread.name} - Initilizing recommendation transfer")

                requestProcessed = False # True if the request was sent or a command to kill the active request was recieved
                while not(requestProcessed or exitFlag):

                    with dsThread._karLock:
                        if (dsThread._killActiveRequest): 
                            logger.debug(f"==={dsThread.name} - Terminating active recommendation transfer")
                            requestProcessed = True 
                            dsThread._killActiveRequest = False
                        else:
                            logger.debug(f"==={dsThread.name} - Successful recommendation transfer") 
                            requestProcessed = dsThread.ai.transfer_rec(data)

                    if not(requestProcessed): 
                        logger.debug(f"==={dsThread.name} - Waiting to resend recommendation")
                        time.sleep(2)


                if(dsThread.workQueue.qsize() % 10 == 0):
                    logger.debug(f"{dsThread.workQueue.qsize()} jobs in queue for {dsThread.name}")

            finally:
                dsThread.workQueue.task_done()
                with dsThread._prLock:
                    dsThread._processingRequest = False
                

        except queue.Empty:
            pass   



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

def calculate_version():
    initpy = open('ai/_version.py').read().split('\n')
    version = list(filter(lambda x: '__version__' in x, initpy))[0].split('\'')[1]
    return version

package_version = calculate_version()

setup(
    name='pennaipy',
    version=package_version,
    author='Heather Williams, Weixuan Fu, William La Cava',
    author_email='hwilli@pennmedicine.upenn.edu, weixuanf@pennmedicine.upenn.edu, lacava@upenn.edu',
    packages=find_packages(exclude=("tests",)),
    url='https://github.com/epistasislab/pennai',
    license='GNU/GPLv3',
    test_suite='nose.collector',
    tests_require=['nose'],
    description=('Penn Artificial Intelligence Data Assistant'),
    long_description='''
A system that intelligently manages machine learning workflows for data science

Contact:
===
- Heather Williams (hwilli@upenn.edu)
- Weixuan Fu (weixuanf@pennmedicine.upenn.edu)
- William La Cava (lacava@upenn.edu)

This project is hosted at https://github.com/epistasislab/pennai
''',
    zip_safe=True,
    install_requires=['numpy>=1.16.3',
                    'scipy>=1.3.1',
                    'scikit-learn>=0.22.0',
                    'update_checker>=0.16',
                    'pandas>=0.24.2',
                    'joblib>=0.13.2'
                    ],
    classifiers=[
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering :: Artificial Intelligence'
    ],
    keywords=['data science', 'machine learning','metalearning'],
)



classifier_config_dict = {

    'sklearn.tree.DecisionTreeClassifier': {
        'params': {
            'criterion': ["gini", "entropy"],
            'max_depth': [3, 5, 10],
            'min_samples_split': [2, 5, 10, 20],
            'min_samples_leaf': [1, 5, 10, 20],
            'min_weight_fraction_leaf': [0.0, 0.05, 0.1, 0.15, 0.2,  0.25, 0.3, 0.35, 0.4, 0.45],
            'max_features': ["sqrt", "log2", None]
        }

    },

    'sklearn.ensemble.GradientBoostingClassifier': {
        'params': {
            'n_estimators': [100, 500],
            'learning_rate': [0.01, 0.1, 1],
            'max_depth': [1, 3, 5, 10],
            'min_samples_split': [2, 5, 10, 20],
            'min_samples_leaf': [1, 5, 10, 20],
            'subsample': [0.5, 1],
            'max_features': ["sqrt", "log2"]
        }
    },

    'sklearn.neighbors.KNeighborsClassifier': {
        'params': {
            'n_neighbors': [1, 3, 5, 7, 9, 11],
            'weights': ["uniform", "distance"],
            'p': [1, 2]
        },
    },

    'sklearn.svm.SVC': {
        'params': {
            'kernel': ["poly", "rbf"],
            'tol': [1e-5, 1e-4, 1e-3, 1e-2, 1e-1],
            'C': [0.0001, 0.001, 0.01, 0.1, 1, 10, 100],
            'gamma': [0.0001, 0.001, 0.01, 0.1, 1, 10, 100],
            'degree': [2, 3],
            'coef0': [0.0, 0.0001, 0.001, 0.01, 0.1, 1, 10]
        },
        'static_parameters': {
            'cache_size': 700, # static_parameters
            'max_iter': 10000, # static_parameters
            'probability': True, # static_parameters
        }
    },

    'sklearn.linear_model.LogisticRegression': {
        'params': {
            'penalty': ["l1", "l2"],
            'C': [0.0001, 0.001, 0.01, 0.1, 1, 10, 100],
            'dual': [True, False],
            'fit_intercept': [True, False]
        },
        'static_parameters': {
            'solver': "liblinear",
            'multi_class': "auto",
        },
        #invalidParameterCombinations
        "invalid_params_comb" : [
            [{"penalty":"l1"}, {"dual": True}]
        ]
    },

    'sklearn.ensemble.RandomForestClassifier': {
        'params': {
            'n_estimators': [100, 500],
            'criterion': ["gini", "entropy"],
            'max_features': ["sqrt", "log2", None],
            'min_samples_split': [2, 5, 10, 20],
            'min_samples_leaf':  [1, 5, 10, 20],
            'bootstrap': [True, False],
            'min_weight_fraction_leaf': [0.0, 0.05, 0.1, 0.15, 0.2, 0.25, 0.3, 0.35, 0.4, 0.45]
        }
    }
}



# first, we need to construct a dataset that has each parameter as a feature. to do this we have to deconstruct
#the 'parameter' column into new columns.
import pandas as pd
import dask.dataframe as dd
from tqdm import tqdm
pd.options.mode.chained_assignment = None

# function to check if parameter is a number.
def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        return False
############################################################## data column types
column_types = {
    'fit_intercept':bool,
    'loss':str,
    'C':float,
    'alpha':float,
    'learning_rate':str,
    'penalty':str,
    'l1_ratio':float,
    'eta0':float,
    'power_t':int,
    'min_weight_fraction_leaf':float,
    'criterion':str,
    'n_estimators':int,
    'max_features':str,
    'Unnamed: 19':str,
    'degree':int,
    'kernel':str,
    'gamma':str,
    'coef0':float,
    'dual':bool,
    'weights':str,
    'n_neighbors':int,
    'fit_prior':bool,
    'max_depth':int,
    'binarize':bool,
}


######################################################### get data metafeatures
# metafeatures = pd.read_csv('data_metafeatures.csv',sep=',',index_col=0)
# print('loaded ',
#       metafeatures.shape[1]-1, ' metafeatures for ',
#       metafeatures.shape[0], ' datasets')
################################################################# get ML results
print('loading pmlb results data...')
data = pd.read_csv('sklearn-benchmark5-data.tsv.gz', sep='\t',
                   names=['dataset',
                         'classifier',
                         'parameters',
                         'accuracy',
                         'macrof1',
                         'bal_accuracy']).fillna('')
data.describe()


############################################## make typed columns for parameters
print('converting parameters to features...')

for i,p in tqdm(enumerate(data['parameters'][:1000])):
    d = dict()
    for ps in p.split(','):
        if is_number(ps.split('=')[-1]):
            d[str(ps.split('=')[0])] = float(ps.split('=')[-1])
        else:
            d[str(ps.split('=')[0])] = ps.split('=')[-1]

    # add columns to data
    for key, val in d.items():
        if key is not '':
            if key not in data.columns:
                if column_types[key] == str:
                    data[key] = ' '
                    data[key] = data[key].apply(lambda x: str(x))
                elif column_types[key] == float:
                    data[key] = 0.0
                    data[key] = data[key].apply(lambda x: float(x))
                elif column_types[key] == bool:
                    data[key] = False
                    data[key] = data[key].apply(lambda x: bool(x))
                else:
                    data[key] = 0
                    data[key] = data[key].apply(lambda x: int(x))


data.to_csv('sklearn-benchmark5-data-ml-parameters-nan-small.tsv',
            sep='\t',index=False)
################################################################# fill in values
def write_param(x):
    #x: dataframe partition
    #gNB has no parameters so ignore

    for i,row in x.iterrows():
        p = row['parameters']
        # try:
        if row['classifier'] == 'GaussianNB' or type(row['parameters'])==str:
            continue
        # except KeyError:
        #     print('x[''parameters'']:',x['parameters'])
        #     print('i:',i,'p:',p)

        d = dict()
        try:
            for ps in p.split(','):
                if ps.split('=')[0] != '':
                    if is_number(ps.split('=')[-1]):
                        d[str(ps.split('=')[0])] = float(ps.split('=')[-1])
                    else:
                        d[str(ps.split('=')[0])] = ps.split('=')[-1]

                for key, val in d.items():
                    if key not in x.columns:
                        print('key:',key,' not in ',x.columns)
                        continue
                    if key == 'loss':
                        x[key] = x[key].apply(lambda i: str(i))
                    # try:
                    x.set_value(i,key,val)
                    # except ValueError:
                    #     print('key:',key,'val:',val)
                    #     raise ValueError
                    # assert( x[i][key] == val)
        except Exception:
            print('i:',i,'p:',p)
            # print(repr(e))
            raise ValueError
    return x
import pandas as pd
from dask import delayed
import dask.dataframe as dd
print('reading results data...')
pdf = pd.read_csv('sklearn-benchmark5-data-ml-parameters-nan-small.tsv', sep='\t')
df = dd.from_pandas(pdf,npartitions=20)
print('mapping parameters values to data...')
df2 = df.map_partitions(lambda x: write_param(x))

df2.to_csv('sklearn-benchmark5-data-ml-parameters.tsv',
            sep='\t',index=False)



"""
API utility functions for Penn-AI
"""

import pandas as pd
import numpy as np
import pdb
import simplejson as json
import requests
import itertools as it
import logging
import sys
from enum import Enum, auto, unique
from sklearn.model_selection import ParameterGrid # utility for hyperparams


logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)
ch = logging.StreamHandler()
formatter = logging.Formatter('%(module)s: %(levelname)s: %(message)s')
ch.setFormatter(formatter)
logger.addHandler(ch)

@unique
class AI_STATUS(Enum):
    ON = "on"
    OFF = "off"
    FINISHED = "finished"
    REQUESTED = "requested"

@unique
class RECOMMENDER_STATUS(Enum):
    DISABLED = 'disabled'
    INITIALIZING = 'initializing'
    RUNNING = 'running'

class LabApi:
    """Class for communicating with the PennAI server
    """

    def __init__(self, api_path, user, api_key, extra_payload, verbose):
        """
        :param api_path: string - path to the lab api server
        :param extra_payload: dict - any additional payload that needs to be
            specified
        :param api_key: string -
        :param user: string - test user
        :param verbose: Boolean
        """
        self.api_path = api_path
        self.exp_path = '/'.join([self.api_path,'api/experiments'])

        self.projects_path = '/'.join([self.api_path,'api/v1/projects'])
        self.algo_path = '/'.join([self.api_path,'api/projects'])

        self.data_path = '/'.join([self.api_path,'api/datasets'])
        self.status_path = '/'.join([self.api_path,'api/v1/datasets'])
        self.submit_path = '/'.join([self.api_path,'api/userdatasets'])

        self.recommender_path = '/'.join([self.api_path,'api/recommender'])
        
        self.api_key=api_key
        self.user=user
        self.verbose=False
        self.header = {'content-type': 'application/json'}
        # optional extra payloads (e.g. user id) for posting to the api
        self.extra_payload = extra_payload
        # static payload is the payload that is constant for every API post
        # with api key for this host
        self.static_payload = {'apikey':self.api_key}
        # add any extra payload
        self.static_payload.update(extra_payload)


        logger.debug("==LabApi paths:")
        logger.debug("self.api_path: " + self.api_path)

        logger.debug("self.exp_path: " + self.exp_path)

        logger.debug("self.projects_path: " + self.projects_path)
        logger.debug("self.algo_path: " + self.algo_path)

        logger.debug("self.data_path: " + self.data_path)
        logger.debug("self.status_path: " + self.status_path)
        logger.debug("self.submit_path: " + self.submit_path)

        logger.debug("self.recommender_path: " + self.recommender_path)
     
    def set_recommender_status(self, status):
        """Attempt to set the recommender status

        :param status 
        """
        logger.info("set_recommender_status(" + str(status) + ")")

        payload = {'status':status}
        data_submit_path = '/'.join([self.recommender_path, "status"])
        res = self.__request(path=data_submit_path, payload=payload, method="POST")

        data = json.loads(res.text)
        return data

    def launch_experiment(self, algorithmId, payload):
        """Attempt to start a ml experiment.

        :param algorithmId: string -
        :param payload: dict - dictionary describing the ml experiment parameters

        :returns: dict {value(str): status(str)}
        """
        logger.info("launch_experiment(" + str(algorithmId) + ", " + str(payload))
        assert algorithmId
        rec_path = '/'.join([self.projects_path, algorithmId, 'experiment'])

        # 503 code is returned if no capactiy available
        # This is a valid scenario and should not cause an exception
        res = self.__request(path=rec_path, payload=payload,
            headers=self.header,
            acceptableNotOkStatusCodes=[503])

        submitResponses = json.loads(res.text)

        #parse json response into named array
        submitStatus={}
        if len(submitResponses) > 0:
            for submitI in submitResponses:
                submitStatus[submitI] = submitResponses[submitI]

        return submitStatus

    def get_projects(self):
        """Get the descriptions and parameters of the supported ml algorithms

        :return: dict - algorithm descriptions as returned by api/projects
        """
        logger.info("get_projects()")

        res = self.__request(path=self.projects_path)
        data = json.loads(res.text)
        return data


    def get_filtered_datasets(self, payload):
        """Get datasets with filters

        :param payload: dict
            How to filter the results {'ai':['requested','finished']}

        :return: dict
            datasets that pass the payload filter
        """
        logger.info("get_filtered_datasets()")

        res = self.__request(path=self.data_path, payload=payload)
        data = json.loads(res.text)
        return data

    def get_dataset_ai_status(self, datasetId):
        """Get the ai status of a particular dataset
        """
        logger.info(f"get_dataset_ai_status('{datasetId}')")

        payload = {}
        path = '/'.join([self.data_path, datasetId])

        res = self.__request(path=path, payload=payload, method="GET")
        data = json.loads(res.text)
        data = data[0]

        #logger.debug(f"get_dataset_ai_status data: {data}")

        aiStatus = None
        if ("ai" in data):
            aiStatus = data["ai"]

        logger.info(f"get_dataset_ai_status('{datasetId}') = '{aiStatus}'")

        return aiStatus

    def get_new_experiments(self, last_update):
        """Get experiments that occurred after last_update

        :param last_update: int -

        :returns: dict - ml experiments results
        """
        logger.info("get_new_experiments(" + str(last_update)+ ")")

        payload = {'date_start':last_update,'has_scores':True}

        res = self.__request(path=self.exp_path, payload=payload)
        data = json.loads(res.text)
        return data

    def get_new_experiments_as_dataframe(self, last_update):
        """
        Get experiments that occured after last_update and return a dataframe

        :param last_update: int -

        :returns: dataframe - last experiments

        """
        logger.info("get_new_experiments_as_dataframe(" + str(last_update)+ ")")

        data = self.get_new_experiments(last_update)

        processed_data = []
        for d in data:
            if ('_options' in d.keys() and '_scores' in d.keys()
                and '_dataset_id' in d.keys()):
                frame={
                    'dataset_id':d['_dataset_id'],
                    'algorithm_id':d['_project_id'],
                    'parameters':d['_options'],
                    'prediction_type':d['_prediction_type'],
                }

                if(d['_prediction_type'] == "classification"):
                    #! This is balanced accuracy!
                    frame['accuracy'] = d['_scores']['accuracy_score']
                    frame['f1'] = d['_scores']['f1_score']

                elif(d['_prediction_type'] == "regression"):
                    frame['r2_cv_mean'] = d['_scores']['r2_score']
                else:
                    msg = (f'error in get_new_experiments_as_dataframe(), '
                            'experiment has unhandled _prediction_type '
                            f'{d["_prediction_type"]}')
                    raise RuntimeError(msg)
                new_experiment = pd.DataFrame([frame])
                assert not new_experiment.isna().any().any(), \
                        "Nan results in experiment"
                processed_data.append(frame)
            else:
              logger.error("new results are missing these fields:",
                      '_options' if '_options' not in d.keys() else '',
                      '_scores' if '_scores' not in d.keys() else '',
                      '_dataset_id' if '_dataset_id' not in d.keys() else '')

        new_experiments = pd.DataFrame(processed_data)

        return new_experiments

    def set_ai_status(self, datasetId, aiStatus):
        """set the ai status for the given dataset.

        :param datasetId: string - dataset to update
        :param aiStatus: string
        """
        logger.info("set_ai_status(" + str(datasetId) +", " + str(aiStatus) + ")")

        payload = {'ai':aiStatus}
        data_submit_path = '/'.join([self.submit_path, datasetId,'ai'])
        res = self.__request(path=data_submit_path, payload=payload, method="PUT")
        return res

    # @Deprecated, used in ai.py; redundant
    def get_ml_id_dict(self):
        """Returns a dictionary for converting algorithm IDs to names.

        :return: dict {mlId(str):mlName(str)}
        """
        logger.info("get_ml_id_dict()")

        res = self.__request(path=self.algo_path)
        responses = json.loads(res.text)

        ml_id_to_name = {}
        for ml in responses:
            ml_id_to_name.update({ml['_id']:ml['name']})

        return ml_id_to_name


    # @Deprecated, used in ai.py; redundant
    def get_user_datasets(self, user):
        """Returns a dictionary for converting dataset IDs to names.

        :return: dict {datasetId(str):datasetName(str)}
        """
        logger.info("get_user_datasets(" + str(user) + ")")

        payload = {'username':user}
        res = self.__request(path=self.submit_path, payload=payload)

        responses = json.loads(res.text)
        dataset_id_to_name = {}
        for dataset in responses:
            dataset_id_to_name.update({dataset['_id']:dataset['name']})
        return dataset_id_to_name


    def get_metafeatures(self, datasetId):
            """Fetches dataset metafeatures, returning dataframe.

            :param datasetId: dataset ID
            :return df: a dataframe of metafeatures, sorted by mf name
            """
            logger.info("get_metafeatures(" + str(datasetId) + ")")

            try:
                res = self.__request(path=self.data_path+'/'+datasetId,
                        method='GET')
                data = json.loads(res.text)

            except Exception as e:
                logger.error('exception when grabbing metafeature data for'
                        + str(datasetId))
                raise e

            data = data[0]
            mf = [data['metafeatures']]
            df = pd.DataFrame.from_records(mf,columns=mf[0].keys())
            print('api_utils:get_metafeatures')
            #include dataset name
            df['dataset'] = data['name']
            df.sort_index(axis=1, inplace=True)

            #logger.debug("metafeatures:")
            #logger.debug(df.head())

            return df

    def valid_combo(self, combo, bad_combos):
        """Checks if parameter combination is valid."""
        for bad_combo in bad_combos:
            bc = {}
            for b in bad_combo:
                bc.update(b)
            bad = True
            for k,v in bc.items():
                if combo[k] != v:
                    bad = False
        return not bad

    def get_all_ml_p(self, categoryFilter = None):
        """
        Returns a list of ml and parameter options for the user 'pennai'
        from the server.

        :returns: pd.DataFrame - unique ml algorithm and parameter combinations
            with columns 'algorithm', 'category', 'parameters'
            'parameters' is a dictionary of parameters
        """
        logger.info("get_all_ml_p()")
        payload = {"username":"pennai"}

        # get the algorithm list for the user 'pennai'
        r = self.__request(path=self.api_path+'/api/preferences', payload=payload,
                method='GET')
        response = json.loads(r.text)

        if len(response) != 1:
            msg = 'error: get_all_ml_p() got ' + str(len(response))
            + ' user preferences, expected 1.'
            logger.error(msg)
            logger.error(response)
            raise RuntimeError(msg)

        if (response[0]['username'] != 'pennai'):
            msg = ('error: get_all_ml_p() did not get user "pennai", got "'
                    + str(response[0]['username']) + '"')
            logger.error(msg)
            logger.error(response)
            raise RuntimeError(msg)

        algorithms = response[0]['algorithms']
        logger.debug('response.algorithms length(): ' + str(len(algorithms)))

        if (categoryFilter):
            assert isinstance(categoryFilter, str)
            algorithms = [a for a in algorithms if a['category'] == categoryFilter]

        result = [] # returned value
        good_def = True # checks that json for ML is in good form

        for i,x in enumerate(algorithms):
            logger.debug('Checking ML: ' + str(x['name']))
            hyperparams = x['schema'].keys()
            hyperparam_dict = {}

            # get a dictionary of hyperparameters and their values
            for h in hyperparams:
                logger.debug('  Checking hyperparams: x[''schema''][h]' +
                        str(x['schema'][h]))
                if 'ui' in x['schema'][h]:
                    if 'values' in x['schema'][h]['ui']:
                        hyperparam_dict[h] = x['schema'][h]['ui']['values']
                    else:
                        hyperparam_dict[h] = x['schema'][h]['ui']['choices']
                else:
                    good_def = False
            if good_def:
                all_hyperparam_combos = list(ParameterGrid(hyperparam_dict))
                #print('\thyperparams: ',hyperparam_dict)
                print(len(all_hyperparam_combos),'hyperparameter combinations for',
                        x['name'])

                for ahc in all_hyperparam_combos:
                    if 'invalidParameterCombinations' in x.keys():
                        if not self.valid_combo(ahc,
                                x['invalidParameterCombinations']):
                            continue
                    result.append({'algorithm':x['name'],
                                   'category':x['category'],
                                   'parameters':ahc,
                                   'algorithm_id':x['_id']})
            else:
                logger.error('warning: ' + str(x['name']) + 'was skipped')
            good_def = True


        # convert to dataframe, making sure there are no duplicates
        all_ml_p = pd.DataFrame(result)
        tmp = all_ml_p.copy()
        tmp['parameters'] = tmp['parameters'].apply(str)
        assert ( len(all_ml_p) == len(tmp.drop_duplicates()) )

        if (len(all_ml_p) > 0):
            logger.info(str(len(all_ml_p)) + ' ml-parameter options loaded')
            logger.info('get_all_ml_p() algorithms:' + str(all_ml_p.algorithm.unique()))
        else:
            logger.error('get_all_ml_p() parsed no results')

        return all_ml_p


    def __request(self, path, payload = None, method = 'POST',
            headers = {'content-type': 'application/json'},
            acceptableNotOkStatusCodes = []):
        """
        Attempt to make an api request and return the result.
        Throw an exception if the request fails or if a status code >400 is returned.

        :return: Requests.response object
        """

        logger.debug("Starting LabApi.__request(" + str(path) + ", " +
                str(payload) + ", " + str(method) + ", ...)" )

        if payload:
            assert isinstance(payload, dict)
            payload.update(self.static_payload)
        else:
            payload = self.static_payload

        res = None
        try:
            res = requests.request(method, path,
                    data=json.dumps(payload, ignore_nan=True),
                    headers=headers)
        except:
            logger.error("Unexpected error in LabApi.__request for path '" +
                    str(method) + ":" + str(path) + "':" + str(sys.exc_info()[0]))
            raise

        # check the status code
        if ((not res.ok)
            and (res.status_code not in acceptableNotOkStatusCodes)):
            msg = ("Request " + str(method) + " status_code not ok, path: '" +
                    str(path) + "' status code: '" + str(res.status_code) +
                    "'' response text: " + str(res.text))
            logger.error(msg)
            raise RuntimeError(msg)


        #logger.debug("Got response LabApi.__request(" + str(path) + ", ..., " + str(method) + ", ...)" )
        # try:
        #     logger.debug("response: ", str(res.text), "\n")
        # except:
        #     logger.debug("couldn't text parse response")
        return res
